import tensorflow as tf
from tensorflow.keras.layers import Layer


class GetDropConnect(Layer):
    """Dropout for model layers.
    DropConnect is similar to dropout, but instead of setting
    activations to zero, it sets a fraction of the weights in a layer to
    zero. This helps to prevent overfitting by reducing the complexity
    of the model and encouraging the model to rely on a more diverse set
    of weights.

    # Arguments
        survival_rate: Float, survival probability to drop features.

    # Properties
        survival_rate: Float.

    # Methods
        call()

    # References
        [Deep Networks with Stochastic Depth]
        (https://arxiv.org/pdf/1603.09382.pdf)
    """
    def __init__(self, survival_rate, **kwargs):
        super(GetDropConnect, self).__init__(**kwargs)
        self.survival_rate = survival_rate

    def call(self, features, training=None):
        if training:
            batch_size = tf.shape(features)[0]
            random_tensor = self.survival_rate
            kwargs = {"shape": [batch_size, 1, 1, 1], "dtype": features.dtype}
            random_tensor = random_tensor + tf.random.uniform(**kwargs)
            binary_tensor = tf.floor(random_tensor)
            output = (features / self.survival_rate) * binary_tensor
            return output
        else:
            return features


class FuseFeature(Layer):
    """Fuse features from different resolutions and return a
    weighted sum. The resulting weighted sum is the fused feature.
    Lower layers of the network tend to extract more basic features,
    such as edges and shapes, while higher layers extract more complex
    features that are useful for making predictions.
    This class implements function that combines features from various
    levels/layers of the model. This helps to combine the strengths of
    different features and create a more robust and accurate
    representation of the input image.

    # Arguments
        fusion: Str, feature fusion method.

    # Properties
        fusion: Str.

    # Methods
        build()
        call()
        _fuse_fast()
        _fuse_sum()
        get_config()

    # References
        [EfficientDet: Scalable and Efficient Object Detection]
        (https://arxiv.org/pdf/1911.09070.pdf)
    """
    def __init__(self, fusion, **kwargs):
        super().__init__(**kwargs)
        self.fusion = fusion
        if fusion == 'fast':
            self.fuse_method = self._fuse_fast
        elif fusion == 'sum':
            self.fuse_method = self._fuse_sum
        else:
            raise ValueError('FPN weight fusion is not defined')

    def build(self, input_shape):
        num_in = len(input_shape)
        args = (self.name, (num_in,), tf.float32,
                tf.keras.initializers.constant(1 / num_in))
        self.w = self.add_weight(*args, trainable=True)

    def call(self, inputs, fusion):
        inputs = [input for input in inputs if input is not None]
        return self.fuse_method(inputs)

    def _fuse_fast(self, inputs):
        w = tf.keras.activations.relu(self.w)

        pre_activations = []
        for input_arg in range(len(inputs)):
            pre_activations.append(w[input_arg] * inputs[input_arg])
        x = tf.reduce_sum(pre_activations, 0)
        x = x / (tf.reduce_sum(w) + 0.0001)
        return x

    def _fuse_sum(self, inputs):
        x = inputs[0]
        for node in inputs[1:]:
            x = x + node
        return x

    def get_config(self):
        config = super().get_config().copy()
        config.update({'fusion': self.fusion})
        return config



from tensorflow.keras.layers import Conv2D, BatchNormalization, SeparableConv2D
from tensorflow.keras.layers import Activation, MaxPooling2D, Add, Input
from tensorflow.keras.layers import GlobalAveragePooling2D
from tensorflow.keras import Model
from tensorflow.keras.regularizers import l2
from tensorflow.keras.models import load_model
from tensorflow.keras.utils import get_file


URL = 'https://github.com/oarriaga/altamira-data/releases/download/v0.6/'


def xception_block(input_tensor, num_kernels, l2_reg=0.01):
    """Xception core block.

    # Arguments
        input_tenso: Keras tensor.
        num_kernels: Int. Number of convolutional kernels in block.
        l2_reg: Float. l2 regression.

    # Returns
        output tensor for the block.
    """
    residual = Conv2D(num_kernels, 1, strides=(2, 2),
                      padding='same', use_bias=False)(input_tensor)
    residual = BatchNormalization()(residual)
    x = SeparableConv2D(
        num_kernels, 3, padding='same',
        kernel_regularizer=l2(l2_reg), use_bias=False)(input_tensor)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = SeparableConv2D(num_kernels, 3, padding='same',
                        kernel_regularizer=l2(l2_reg), use_bias=False)(x)
    x = BatchNormalization()(x)

    x = MaxPooling2D(3, strides=(2, 2), padding='same')(x)
    x = Add()([x, residual])
    return x


def build_xception(
        input_shape, num_classes, stem_kernels, block_kernels, l2_reg=0.01):
    """Function for instantiating an Xception model.

    # Arguments
        input_shape: List corresponding to the input shape of the model.
        num_classes: Integer.
        stem_kernels: List of integers. Each element of the list indicates
            the number of kernels used as stem blocks.
        block_kernels: List of integers. Each element of the list Indicates
            the number of kernels used in the xception blocks.
        l2_reg. Float. L2 regularization used in the convolutional kernels.

    # Returns
        Tensorflow-Keras model.

    # References
        - [Xception: Deep Learning with Depthwise Separable
            Convolutions](https://arxiv.org/abs/1610.02357)
    """

    x = inputs = Input(input_shape, name='image')
    for num_kernels in stem_kernels:
        x = Conv2D(num_kernels, 3, kernel_regularizer=l2(l2_reg),
                   use_bias=False, padding='same')(x)
        x = BatchNormalization()(x)
        x = Activation('relu')(x)

    for num_kernels in block_kernels:
        x = xception_block(x, num_kernels, l2_reg)

    x = Conv2D(num_classes, 3, kernel_regularizer=l2(l2_reg),
               padding='same')(x)
    # x = BatchNormalization()(x)
    x = GlobalAveragePooling2D()(x)
    output = Activation('softmax', name='label')(x)

    model_name = '-'.join(['XCEPTION',
                           str(input_shape[0]),
                           str(stem_kernels[0]),
                           str(len(block_kernels))
                           ])
    model = Model(inputs, output, name=model_name)
    return model


def MiniXception(input_shape, num_classes, weights=None):
    """Build MiniXception (see references).

    # Arguments
        input_shape: List of three integers e.g. ``[H, W, 3]``
        num_classes: Int.
        weights: ``None`` or string with pre-trained dataset. Valid datasets
            include only ``FER``.

    # Returns
        Tensorflow-Keras model.

    # References
       - [Real-time Convolutional Neural Networks for Emotion and
            Gender Classification](https://arxiv.org/abs/1710.07557)
    """
    if weights == 'FER':
        filename = 'fer2013_mini_XCEPTION.119-0.65.hdf5'
        path = get_file(filename, URL + filename, cache_subdir='paz/models')
        model = load_model(path)
    else:
        stem_kernels = [32, 64]
        block_data = [128, 128, 256, 256, 512, 512, 1024]
        model_inputs = (input_shape, num_classes, stem_kernels, block_data)
        model = build_xception(*model_inputs)
    model._name = 'MINI-XCEPTION'
    return model




import os
import tensorflow as tf
from tensorflow.keras.utils import get_file
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.initializers import truncated_normal
from tensorflow.keras.regularizers import l2
from tensorflow.keras.layers import Input, Activation, Reshape
from tensorflow.keras.models import Model


WEIGHT_PATH = ('https://github.com/oarriaga/altamira-data/releases/download/'
               'v0.14/iknet_weight.hdf5')


def dense(x, num_units):
    x = Dense(num_units, activation=None, kernel_regularizer=l2(0.5 * 1.0),
              kernel_initializer=truncated_normal(stddev=0.01))(x)
    return x


def block(x, num_units):
    x = dense(x, num_units)
    x = BatchNormalization()(x)
    return x


def normalize(x):
    norm = tf.norm(x, axis=-1, keepdims=True)
    norm = tf.maximum(norm, 1e-6)
    normalized_x = x / norm
    return normalized_x


def reorder_quaternions(quaternions):
    w = quaternions[:, :, 0:1]
    qs = quaternions[:, :, 1:4]
    quaternions = tf.concat((qs, w), axis=-1)
    return quaternions


def IKNet(input_shape=(84, 3), num_keypoints=21, depth=6, width=1024):
    """IKNet: Estimate absolute joint angle for the minimal hand keypoints.

    # Arguments
        input_shape: [num_keypoint x 4, 3]. Contains 3D keypoints, bone
                     orientation, refrence keypoint, refrence bone orientation.
        num_keypoints: Int. Number of keypoints.

    # Returns
        Tensorflow-Keras model.
        absolute joint angle in quaternion representation.

    # Reference
        - [Monocular Real-time Hand Shape and Motion Capture using Multi-modal
           Data](https://arxiv.org/abs/2003.09572)
    """
    input = Input(shape=input_shape, dtype=tf.float32)
    x = Reshape([1, -1])(input)

    for depth_arg in range(depth):
        x = block(x, width)
        x = Activation('sigmoid')(x)
    x = dense(x, num_keypoints * 4)
    x = Reshape([num_keypoints, 4])(x)
    x = normalize(x)

    positive_mask = tf.tile(x[:, :, 0:1] > 0, [1, 1, 4])
    quaternions = tf.where(positive_mask, x, -x)
    quaternions = reorder_quaternions(quaternions)

    model = Model(input, outputs=[quaternions])

    URL = ('https://github.com/oarriaga/altamira-data/releases/download/'
           'v0.14/iknet_weight.hdf5')
    filename = os.path.basename(URL)
    weights_path = get_file(filename, URL, cache_subdir='paz/models')
    print('==> Loading %s model weights' % weights_path)
    model.load_weights(weights_path)
    return model



from .anchors import build_anchors
from .anchors import build_octaves
from .anchors import build_aspect
from .anchors import build_scales
from .anchors import build_strides
from .anchors import make_branch_boxes
from .anchors import compute_box_coordinates
from .anchors import build_base_anchor
from .anchors import compute_aspect_size
from .anchors import compute_anchor_dims
from .anchors import compute_anchor_centres



import numpy as np
import tensorflow.keras.backend as K


class Projector(object):
    """Projects keypoints from image coordinates to 3D space and viceversa.
    This model uses the camera focal length and the depth estimation of a point
    to project it to image coordinates. It works with numpy matrices or
    tensorflow values. See ``use_numpy``.

    # Arguments
        focal_length: Float. Focal length of camera used to generate keypoints.
        use_numpy: Boolean. If `True` both unproject and project functions
            take numpy arrays as inputs. If `False` takes tf.tensors as inputs.
    """
    def __init__(self, focal_length, use_numpy=False):
        self.focal_length = focal_length
        self.project = self._project_keras
        self.unproject = self._unproject_keras
        if use_numpy:
            self.project = self._project_numpy
            self.unproject = self._unproject_numpy

    def _project_keras(self, xyzw):
        z = xyzw[:, :, 2:3] + 1e-8
        x = - (self.focal_length / z) * xyzw[:, :, 0:1]
        y = - (self.focal_length / z) * xyzw[:, :, 1:2]
        return K.concatenate([x, y, z], axis=2)

    def _project_numpy(self, xyzw):
        z = xyzw[:, :, 2:3] + 1e-8
        x = - (self.focal_length / z) * xyzw[:, :, 0:1]
        y = - (self.focal_length / z) * xyzw[:, :, 1:2]
        return np.concatenate([x, y, z], axis=2)

    def _unproject_keras(self, xyz):
        z = xyz[:, :, 2:3]
        x = - (z / self.focal_length) * xyz[:, :, 0:1]
        y = - (z / self.focal_length) * xyz[:, :, 1:2]
        w = K.ones_like(z)
        xyzw = K.concatenate([x, y, z, w], axis=2)
        return xyzw

    def _unproject_numpy(self, xyz):
        z = xyz[:, :, 2:3]
        x = - (z / self.focal_length) * xyz[:, :, 0:1]
        y = - (z / self.focal_length) * xyz[:, :, 1:2]
        w = np.ones_like(z)
        xyzw = np.concatenate([x, y, z, w], axis=2)
        return xyzw



import numpy as np

# TODO: rmeove cv2 import by computing explicitly rodriguez formula
import cv2


def rotation_vector_to_rotation_matrix(rotation_vector):
    """Transforms rotation vector (axis-angle) form to rotation matrix.

    # Arguments
        rotation_vector: Array (3). Rotation vector in axis-angle form.

    # Returns
        Array (3, 3) rotation matrix.
    """
    rotation_matrix = np.eye(3)
    cv2.Rodrigues(rotation_vector, rotation_matrix)
    return rotation_matrix


def build_rotation_matrix_z(angle):
    """Builds rotation matrix in Z axis.

    # Arguments
        angle: Float. Angle in radians.

    # Return
        Array (3, 3) rotation matrix in Z axis.
    """
    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    rotation_matrix_z = np.array([[+cos_angle, -sin_angle, 0.0],
                                  [+sin_angle, +cos_angle, 0.0],
                                  [0.0, 0.0, 1.0]])
    return rotation_matrix_z


def build_rotation_matrix_x(angle):
    """Builds rotation matrix in X axis.

    # Arguments
        angle: Float. Angle in radians.

    # Return
        Array (3, 3) rotation matrix in Z axis.
    """
    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    rotation_matrix_x = np.array([[1.0, 0.0, 0.0],
                                  [0.0, +cos_angle, -sin_angle],
                                  [0.0, +sin_angle, +cos_angle]])
    return rotation_matrix_x


def build_rotation_matrix_y(angle):
    """Builds rotation matrix in Y axis.

    # Arguments
        angle: Float. Angle in radians.

    # Return
        Array (3, 3) rotation matrix in Z axis.
    """
    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    rotation_matrix_y = np.array([[+cos_angle, 0.0, +sin_angle],
                                  [0.0, 1.0, 0.0],
                                  [-sin_angle, 0.0, +cos_angle]])
    return rotation_matrix_y


def compute_norm_SO3(rotation_mesh, rotation):
    """Computes norm between SO3 elements.

    # Arguments
        rotation_mesh: Array (3, 3), rotation matrix.
        rotation: Array (3, 3), rotation matrix.

    # Returns
        Scalar representing the distance between both rotation matrices.
    """
    difference = np.dot(np.linalg.inv(rotation), rotation_mesh) - np.eye(3)
    distance = np.linalg.norm(difference, ord='fro')
    return distance


def calculate_canonical_rotation(rotation_mesh, rotations):
    """Returns the rotation matrix closest to rotation mesh.

    # Arguments
        rotation_mesh: Array (3, 3), rotation matrix.
        rotations: List of array of (3, 3), rotation matrices.

    # Returns
        Element of list closest to rotation mesh.
    """
    norms = [compute_norm_SO3(rotation_mesh, R) for R in rotations]
    closest_rotation_arg = np.argmin(norms)
    closest_rotation = rotations[closest_rotation_arg]
    canonical_rotation = np.linalg.inv(closest_rotation)
    return canonical_rotation


def rotation_matrix_to_axis_angle(rotation_matrix):
    """Transforms rotation matrix to axis angle.

    # Arguments
        Rotation matrix [3, 3].

    # Returns
        axis_angle: Array containing axis angle represent [wx, wy, wz, theta].
    """
    cos_theta = (np.trace(rotation_matrix) - 1.0) / 2.0
    angle = np.arccos(cos_theta)
    axis = np.array([rotation_matrix[2, 1] - rotation_matrix[1, 2],
                     rotation_matrix[0, 2] - rotation_matrix[2, 0],
                     rotation_matrix[1, 0] - rotation_matrix[0, 1]])
    axis = axis / np.linalg.norm(axis)
    axis_angle = np.hstack([axis, angle])
    return axis_angle


def rotation_matrix_to_compact_axis_angle(matrix):
    """Transforms rotation matrix to compact axis angle.

    # Arguments
        Rotation matrix [3, 3].

    # Returns
        compact axis_angle
    """
    axis_angle = rotation_matrix_to_axis_angle(matrix)
    axis = axis_angle[:3]
    angle = axis_angle[3]
    compact_axis_angle = axis * angle
    return compact_axis_angle



import os
import numpy as np

from tensorflow.keras.callbacks import Callback
import tensorflow.keras.backend as K

from ..backend.image import write_image

from paz.evaluation import evaluateMAP


class DrawInferences(Callback):
    """Saves an image with its corresponding inferences

    # Arguments
        save_path: String. Path in which the images will be saved.
        images: List of numpy arrays of shape.
        pipeline: Function that takes as input an element of ''images''
            and outputs a ''Dict'' with inferences.
        topic: Key to the ''inferences'' dictionary containing as value the
            drawn inferences.
        verbose: Integer. If is bigger than 1 messages would be displayed.
    """
    def __init__(self, save_path, images, pipeline, topic='image', verbose=1):
        super(DrawInferences, self).__init__()
        self.save_path = os.path.join(save_path, 'images')
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        self.pipeline = pipeline
        self.images = images
        self.topic = topic
        self.verbose = verbose

    def on_epoch_end(self, epoch, logs=None):
        for image_arg, image in enumerate(self.images):
            inferences = self.pipeline(image.copy())
            epoch_name = 'epoch_%03d' % epoch
            save_path = os.path.join(self.save_path, epoch_name)
            if not os.path.exists(save_path):
                os.makedirs(save_path)
            image_name = 'image_%03d.png' % image_arg
            image_name = os.path.join(save_path, image_name)
            write_image(image_name, inferences[self.topic])
        if self.verbose:
            print('Saving predicted images in:', self.save_path)


class LearningRateScheduler(Callback):
    """ Callback for reducing learning rate at specific epochs.

    # Arguments
        learning_rate: float. Indicates the starting learning rate.
        gamma_decay: float. In an scheduled epoch the learning rate
            is multiplied by this factor.
        scheduled_epochs: List of integers. Indicates in which epochs
            the learning rate will be multiplied by the gamma decay factor.
        verbose: Integer. If is bigger than 1 messages would be displayed.
    """
    def __init__(
            self, learning_rate, gamma_decay, scheduled_epochs, verbose=1):
        super(LearningRateScheduler, self).__init__()
        self.learning_rate = learning_rate
        self.gamma_decay = gamma_decay
        self.scheduled_epochs = scheduled_epochs
        self.verbose = verbose

    def on_epoch_begin(self, epoch, logs=None):
        if not hasattr(self.model.optimizer, 'lr'):
            raise ValueError('Optimizer must have a "lr" attribute.')

        learning_rate = float(K.get_value(self.model.optimizer.lr))
        learning_rate = self.schedule(epoch)
        if not isinstance(learning_rate, (float, np.float32, np.float64)):
            raise ValueError('Learning rate should be float.')
        K.set_value(self.model.optimizer.lr, learning_rate)
        if self.verbose > 0:
            print('\nEpoch %05d: LearningRateScheduler reducing learning '
                  'rate to %s.' % (epoch + 1, learning_rate))

    def schedule(self, epoch):
        if epoch in self.scheduled_epochs:
            self.learning_rate = self.learning_rate * self.gamma_decay
        return self.learning_rate


class EvaluateMAP(Callback):
    """Evaluates mean average precision (MAP) of an object detector.

    # Arguments
        data_manager: Data manager and loader class. See ''paz.datasets''
            for examples.
        detector: Tensorflow-Keras model.
        period: Int. Indicates how often the evaluation is performed.
        save_path: Str.
        iou_thresh: Float.
    """
    def __init__(
            self, data_manager, detector, period, save_path, iou_thresh=0.5):
        super(EvaluateMAP, self).__init__()
        self.data_manager = data_manager
        self.detector = detector
        self.period = period
        self.save_path = save_path
        self.dataset = data_manager.load_data()
        self.iou_thresh = iou_thresh
        self.class_names = self.data_manager.class_names
        self.class_dict = {}
        for class_arg, class_name in enumerate(self.class_names):
            self.class_dict[class_name] = class_arg

    def on_epoch_end(self, epoch, logs):
        if (epoch + 1) % self.period == 0:
            result = evaluateMAP(
                self.detector,
                self.dataset,
                self.class_dict,
                iou_thresh=self.iou_thresh,
                use_07_metric=True)

            result_str = 'mAP: {:.4f}\n'.format(result['map'])
            metrics = {'mAP': result['map']}
            for arg, ap in enumerate(result['ap']):
                if arg == 0 or np.isnan(ap):  # skip background
                    continue
                metrics[self.class_names[arg]] = ap
                result_str += '{:<16}: {:.4f}\n'.format(
                    self.class_names[arg], ap)
            print(result_str)

            # Saving the evaluation results
            filename = os.path.join(self.save_path, 'MAP_Evaluation_Log.txt')
            with open(filename, 'a') as eval_log_file:
                eval_log_file.write('Epoch: {}\n{}\n'.format(
                    str(epoch), result_str))



import os
from tensorflow.keras.utils import to_categorical
import numpy as np

from .utils import get_class_names
from ..abstract import Loader
from ..backend.image import resize_image


class FER(Loader):
    """Class for loading FER2013 emotion classification dataset.
    # Arguments
        path: String. Full path to fer2013.csv file.
        split: String. Valid option contain 'train', 'val' or 'test'.
        class_names: String or list: If 'all' then it loads all default
            class names.
        image_size: List of length two. Indicates the shape in which
            the image will be resized.

    # References
        -[FER2013 Dataset and Challenge](kaggle.com/c/challenges-in-\
            representation-learning-facial-expression-recognition-challenge)
    """

    def __init__(
            self, path, split='train', class_names='all', image_size=(48, 48)):

        if class_names == 'all':
            class_names = get_class_names('FER')

        path = os.path.join(path, 'fer2013.csv')
        super(FER, self).__init__(path, split, class_names, 'FER')
        self.image_size = image_size
        self._split_to_filter = {'train': 'Training', 'val': 'PublicTest',
                                 'test': 'PrivateTest'}

    def load_data(self):
        data = np.genfromtxt(self.path, str, delimiter=',', skip_header=1)
        data = data[data[:, -1] == self._split_to_filter[self.split]]
        faces = np.zeros((len(data), *self.image_size))
        for sample_arg, sample in enumerate(data):
            face = np.array(sample[1].split(' '), dtype=int).reshape(48, 48)
            face = resize_image(face, self.image_size)
            faces[sample_arg, :, :] = face
        emotions = to_categorical(data[:, 0].astype(int), self.num_classes)

        data = []
        for face, emotion in zip(faces, emotions):
            sample = {'image': face, 'label': emotion}
            data.append(sample)
        return data



from __future__ import division

import numpy as np

from ..abstract import Processor, Box2D
from ..backend.boxes import match
from ..backend.boxes import encode
from ..backend.boxes import decode
from ..backend.boxes import offset
from ..backend.boxes import clip
from ..backend.boxes import nms_per_class
from ..backend.boxes import merge_nms_box_with_class
from ..backend.boxes import denormalize_box
from ..backend.boxes import make_box_square
from ..backend.boxes import filter_boxes
from ..backend.boxes import scale_box


class SquareBoxes2D(Processor):
    """Transforms bounding rectangular boxes into square bounding boxes.
    """
    def __init__(self):
        super(SquareBoxes2D, self).__init__()

    def call(self, boxes2D):
        for box2D in boxes2D:
            box2D.coordinates = make_box_square(box2D.coordinates)
        return boxes2D


class DenormalizeBoxes2D(Processor):
    """Denormalizes boxes shapes to be in accordance to the original
    image size.

    # Arguments:
        image_size: List containing height and width of an image.
    """
    def __init__(self):
        super(DenormalizeBoxes2D, self).__init__()

    def call(self, image, boxes2D):
        shape = image.shape[:2]
        for box2D in boxes2D:
            box2D.coordinates = denormalize_box(box2D.coordinates, shape)
        return boxes2D


class RoundBoxes2D(Processor):
    """Round to integer box coordinates.
    """
    def __init__(self):
        super(RoundBoxes2D, self).__init__()

    def call(self, boxes2D):
        for box2D in boxes2D:
            box2D.coordinates = [int(x) for x in box2D.coordinates]
        return boxes2D


class FilterClassBoxes2D(Processor):
    """Filters boxes with valid class names.

    # Arguments
        valid_class_names: List of strings indicating class names to be kept.
    """
    def __init__(self, valid_class_names):
        self.valid_class_names = valid_class_names
        super(FilterClassBoxes2D, self).__init__()

    def call(self, boxes2D):
        filtered_boxes2D = []
        for box2D in boxes2D:
            if box2D.class_name in self.valid_class_names:
                filtered_boxes2D.append(box2D)
        return filtered_boxes2D


class CropBoxes2D(Processor):
    """Creates a list of images cropped from the bounding boxes.

    # Arguments
        offset_scales: List of floats having x and y scales respectively.
    """
    def __init__(self):
        super(CropBoxes2D, self).__init__()

    def call(self, image, boxes2D):
        image_crops = []
        for box2D in boxes2D:
            x_min, y_min, x_max, y_max = box2D.coordinates
            image_crops.append(image[y_min:y_max, x_min:x_max])
        return image_crops


class ClipBoxes2D(Processor):
    """Clips boxes coordinates into the image dimensions"""
    def __init__(self):
        super(ClipBoxes2D, self).__init__()

    def call(self, image, boxes2D):
        image_height, image_width = image.shape[:2]
        for box2D in boxes2D:
            box2D.coordinates = clip(box2D.coordinates, image.shape[:2])
        return boxes2D


class OffsetBoxes2D(Processor):
    """Offsets the height and widht of a list of ``Boxes2D``.

    # Arguments
        offsets: Float between [0, 1].
    """
    def __init__(self, offsets):
        super(OffsetBoxes2D, self).__init__()
        self.offsets = offsets

    def call(self, boxes2D):
        for box2D in boxes2D:
            box2D.coordinates = offset(box2D.coordinates, self.offsets)
        return boxes2D


class ToBoxes2D(Processor):
    """Transforms boxes from dataset into `Boxes2D` messages.

    # Arguments
        class_names: List of class names ordered with respect to the
            class indices from the dataset ``boxes``.
        one_hot_encoded: Bool, indicating if scores are one hot vectors.
        default_score: Float, score to set.
        default_class: Str, class to set.
        box_method: Int, method to convert boxes to ``Boxes2D``.

    # Properties
        one_hot_encoded: Bool.
        box_processor: Callable.

    # Methods
        call()
    """
    def __init__(
            self, class_names=None, one_hot_encoded=False,
            default_score=1.0, default_class=None, box_method=0):
        if class_names is not None:
            arg_to_class = dict(zip(range(len(class_names)), class_names))
        self.one_hot_encoded = one_hot_encoded
        method_to_processor = {
            0: BoxesWithOneHotVectorsToBoxes2D(arg_to_class),
            1: BoxesToBoxes2D(default_score, default_class),
            2: BoxesWithClassArgToBoxes2D(arg_to_class, default_score)}
        self.box_processor = method_to_processor[box_method]
        super(ToBoxes2D, self).__init__()

    def call(self, box_data):
        return self.box_processor(box_data)


class BoxesToBoxes2D(Processor):
    """Transforms boxes from dataset into `Boxes2D` messages given no
    class names and score.

    # Arguments
        default_score: Float, score to set.
        default_class: Str, class to set.

    # Properties
        default_score: Float.
        default_class: Str.

    # Methods
        call()
    """
    def __init__(self, default_score=1.0, default_class=None):
        self.default_score = default_score
        self.default_class = default_class
        super(BoxesToBoxes2D, self).__init__()

    def call(self, box_data):
        boxes2D = []
        for box in box_data:
            boxes2D.append(
                Box2D(box[:4], self.default_score, self.default_class))
        return boxes2D


class BoxesWithOneHotVectorsToBoxes2D(Processor):
    """Transforms boxes from dataset into `Boxes2D` messages given boxes
    with scores as one hot vectors.

    # Arguments
        arg_to_class: List, of classes.

    # Properties
        arg_to_class: List.

    # Methods
        call()
    """
    def __init__(self, arg_to_class):
        self.arg_to_class = arg_to_class
        super(BoxesWithOneHotVectorsToBoxes2D, self).__init__()

    def call(self, box_data):
        boxes2D = []
        for box in box_data:
            class_scores = box[4:]
            class_arg = np.argmax(class_scores)
            score = class_scores[class_arg]
            class_name = self.arg_to_class[class_arg]
            boxes2D.append(Box2D(box[:4], score, class_name))
        return boxes2D


class BoxesWithClassArgToBoxes2D(Processor):
    """Transforms boxes from dataset into `Boxes2D` messages given boxes
    with class argument.

    # Arguments
        default_score: Float, score to set.
        arg_to_class: List, of classes.

    # Properties
        default_score: Float.
        arg_to_class: List.

    # Methods
        call()
    """
    def __init__(self, arg_to_class, default_score=1.0):
        self.default_score = default_score
        self.arg_to_class = arg_to_class
        super(BoxesWithClassArgToBoxes2D, self).__init__()

    def call(self, box_data):
        boxes2D = []
        for box in box_data:
            class_name = self.arg_to_class[box[-1]]
            boxes2D.append(Box2D(box[:4], self.default_score, class_name))
        return boxes2D


class RoundBoxes(Processor):
    """Rounds the floating value coordinates of the box coordinates
    into integer type.

    # Methods
        call()
    """
    def __init__(self):
        super(RoundBoxes, self).__init__()

    def call(self, boxes2D):
        for box2D in boxes2D:
            box2D.coordinates = box2D.coordinates.astype(int)
        return boxes2D


class MatchBoxes(Processor):
    """Match prior boxes with ground truth boxes.

    # Arguments
        prior_boxes: Numpy array of shape (num_boxes, 4).
        iou: Float in [0, 1]. Intersection over union in which prior boxes
            will be considered positive. A positive box is box with a class
            different than `background`.
        variance: List of two floats.
    """
    def __init__(self, prior_boxes, iou=.5):
        self.prior_boxes = prior_boxes
        self.iou = iou
        super(MatchBoxes, self).__init__()

    def call(self, boxes):
        boxes = match(boxes, self.prior_boxes, self.iou)
        return boxes


class EncodeBoxes(Processor):
    """Encodes bounding boxes.

    # Arguments
        prior_boxes: Numpy array of shape (num_boxes, 4).
        variances: List of two float values.
    """
    def __init__(self, prior_boxes, variances=[0.1, 0.1, 0.2, 0.2]):
        self.prior_boxes = prior_boxes
        self.variances = variances
        super(EncodeBoxes, self).__init__()

    def call(self, boxes):
        encoded_boxes = encode(boxes, self.prior_boxes, self.variances)
        return encoded_boxes


class DecodeBoxes(Processor):
    """Decodes bounding boxes.

    # Arguments
        prior_boxes: Numpy array of shape (num_boxes, 4).
        variances: List of two float values.
    """
    def __init__(self, prior_boxes, variances=[0.1, 0.1, 0.2, 0.2]):
        self.prior_boxes = prior_boxes
        self.variances = variances
        super(DecodeBoxes, self).__init__()

    def call(self, boxes):
        decoded_boxes = decode(boxes, self.prior_boxes, self.variances)
        return decoded_boxes


class NonMaximumSuppressionPerClass(Processor):
    """Applies non maximum suppression per class.

    # Arguments
        nms_thresh: Float between [0, 1].
        epsilon: Float between [0, 1].
    """
    def __init__(self, nms_thresh=.45, epsilon=0.01):
        self.nms_thresh = nms_thresh
        self.epsilon = epsilon
        super(NonMaximumSuppressionPerClass, self).__init__()

    def call(self, box_data):
        box_data, class_labels = nms_per_class(
            box_data, self.nms_thresh, self.epsilon)
        return box_data, class_labels


class MergeNMSBoxWithClass(Processor):
    """Merges box coordinates with their corresponding class
    defined by `class_labels` which is decided by best box geometry
    by non maximum suppression (and not by the best scoring class)
    into a single output.
    """
    def __init__(self):
        super(MergeNMSBoxWithClass, self).__init__()

    def call(self, box_data, class_labels):
        box_data = merge_nms_box_with_class(box_data, class_labels)
        return box_data


class FilterBoxes(Processor):
    """Filters boxes outputted from function ``detect`` as
    ``Box2D`` messages.

    # Arguments
        class_names: List of class names.
        conf_thresh: Float between [0, 1].
    """
    def __init__(self, class_names, conf_thresh=0.5):
        self.class_names = class_names
        self.conf_thresh = conf_thresh
        self.arg_to_class = dict(zip(
            list(range(len(self.class_names))), self.class_names))
        super(FilterBoxes, self).__init__()

    def call(self, box_data):
        box_data = filter_boxes(box_data, self.conf_thresh)
        return box_data


class CropImage(Processor):
    """Crop images using a list of ``box2D``.
    """
    def __init__(self):
        super(CropImage, self).__init__()

    def call(self, image, box2D):
        x_min, y_min, x_max, y_max = box2D.coordinates
        return image[y_min:y_max, x_min:x_max]


class RemoveClass(Processor):
    """Remove a particular class from the pipeline.

    # Arguments
        class_names: List, indicating given class names.
        class_arg: Int, index of the class to be removed.
        renormalize: Bool, if true scores are renormalized.

    # Properties
        class_arg: Int.
        renormalize: Bool

    # Methods
        call()
    """
    def __init__(self, class_names, class_arg=None, renormalize=False):
        self.class_arg = class_arg
        self.renormalize = renormalize
        if class_arg is not None:
            del class_names[class_arg]
        super(RemoveClass, self).__init__()

    def call(self, box_data):
        if not self.renormalize and self.class_arg is not None:
            box_data = np.delete(box_data, 4 + self.class_arg, axis=1)
        elif self.renormalize:
            raise NotImplementedError
        return box_data


class ScaleBox(Processor):
    """Scale box coordinates of the prediction.

    # Arguments
        scales: Array of shape `()`, value to scale boxes.

    # Properties
        scales: Int.

    # Methods
        call()
    """
    def __init__(self):
        super(ScaleBox, self).__init__()

    def call(self, boxes, scales):
        boxes = scale_box(boxes, scales)
        return boxes



import numpy as np
import tensorflow as tf


def append_values(dictionary, lists, keys):
    """Append dictionary values to lists

    # Arguments
        dictionary: dict
        lists: List of lists
        keys: Keys to dictionary values
    """
    if len(keys) != len(lists):
        assert ValueError('keys and lists must have same length')
    for key_arg, key in enumerate(keys):
        lists[key_arg].append(dictionary[key])
    return lists


def append_lists(intro_lists, outro_lists):
    """Appends multiple new values in intro lists to multiple outro lists

    # Arguments
        intro_lists: List of lists
        outro_lists: List of lists

    # Returns
        Lists with new values of intro lists
    """
    for intro_list, outro_list in zip(intro_lists, outro_lists):
        outro_list.append(intro_list)
    return outro_lists


def get_upper_multiple(x, multiple=64):
    """Returns the upper multiple of 'multiple' to the x.

    # Arguments
        x: Int.
        multiple: Int.

    # Returns
        upper multiple. Int.
    """
    x = x + (multiple - 1)
    floor_value = x // multiple
    upper_multiple = floor_value * multiple
    return upper_multiple


def resize_with_same_aspect_ratio(image, input_size, multiple=64):
    """Resize the sort side of the input image to input_size and keep
    the aspect ratio.

    # Arguments
        input_size: Dimension to be resized. Int.
        H: Int.
        W: Int.

    # Returns
        resized H and W.
    """
    H, W = np.sort(image.shape[:2])
    H_resized = int(input_size)
    W_resized = input_size / H
    W_resized = W_resized * W
    W_resized = int(get_upper_multiple(W_resized, multiple))
    size = np.array([W_resized, H_resized])
    return size


def get_transformation_scale(image, size, scaling_factor):
    """Caluclte scale of resized H and W.

    # Arguments
        H: Int.
        H_resized: Int.
        H_resized: Int.
        scaling_factor: Int.

    # Returns
        scaled H and W
    """
    H, W = image.shape[:2]
    H_resized, W_resized = size

    if H < W:
        H_resized, W_resized = W_resized, H_resized
    H, W = np.sort([H, W])

    scale_H = H / scaling_factor
    aspect_ratio = W_resized / H_resized
    scale_W = aspect_ratio * scale_H
    scale = np.array([scale_W, scale_H])
    return scale


def compare_vertical_neighbours(x, y, image, offset=0.25):
    """Compare two vertical neighbors and add an offset to the smaller one.

    # Arguments
        x: Int. x coordinate of pixel to be compared.
        y: Int. y coordinate of pixel to be compared.
        image: Array.
        offset: Float.
    """
    int_x, int_y = int(x), int(y)
    lower_y = min(int_y + 1, image.shape[1] - 1)
    upper_y = max(int_y - 1, 0)
    if image[int_x, lower_y] > image[int_x, upper_y]:
        y = y + offset
    else:
        y = y - offset
    return y


def compare_horizontal_neighbours(x, y, image, offset=0.25):
    """Compare two horizontal neighbors and add an offset to the smaller one.

    # Arguments
        x: Int. x coordinate of pixel to be compared.
        y: Int. y coordinate of pixel to be compared.
        image: Array.
        offset: Float.
    """
    int_x, int_y = int(x), int(y)
    left_x = max(0, int_x - 1)
    right_x = min(int_x + 1, image.shape[0] - 1)
    if image[right_x, int_y] > image[left_x, int_y]:
        x = x + offset
    else:
        x = x - offset
    return x


def get_all_indices_of_array(array):
    """Get all the indices of an array.

    # Arguments
        array: Array

    # Returns
        Array. Array with the indices of the input array
    """
    all_indices = np.ndarray(array.shape)
    all_indices.fill(True)
    all_indices = np.where(all_indices)
    all_indices = np.array(all_indices).T
    print(all_indices)
    return all_indices


def gather_nd(array, indices, axis):
    """Take the value from the input array on the given indices along the
    given axis.

    # Arguments
        array: Array
        indices: list/Array. values to be gathered from
        axis: Int. Axis along which to gather values.

    # Returns
        Array. Gathered values from the input array
    """
    gathered = np.take_along_axis(array, indices, axis=axis)
    return gathered


def calculate_norm(vector, order=None, axis=None):
    """Calculates the norm of vector.

    # Arguments
        x: List of spatial coordinates (x, y, z)
    """
    return np.linalg.norm(vector, ord=order, axis=axis)


def tensor_to_numpy(tensor):
    """Convert a tensor to a Array.

    # Arguments
        tensor: multidimensional array of type tensor
    """
    return tensor.cpu().numpy()


def pad_matrix(matrix, pool_size=(3, 3), strides=(1, 1),
               padding='valid', value=0):
    """Pad an array

    # Arguments
        matrix: Array.
        padding: String. Type of padding
        value: Int. Value to be added in the padded area.
        poolsize: Int. How many rows and colums to be padded for 'same' padding
    """
    matrix = np.array(matrix)
    H, W = matrix.shape[:2]
    if padding == 'valid':
        padding = ((0, 0), (0, 0))
    if padding == 'square':
        if H > W:
            padding = ((0, 0), (0, H - W))
        else:
            padding = ((0, W - H), (0, 0))
    if padding == 'same':
        if isinstance(pool_size, int):
            pool_size = (pool_size, pool_size)
        if isinstance(strides, int):
            strides = (strides, strides)
        if H % strides[0] == 0:
            height_pad = np.max((pool_size[0] - strides[0]), 0)
        else:
            height_pad = np.max(pool_size[0] - (H % strides[0]), 0)
        if W % strides[1] == 0:
            width_pad = np.max((pool_size[1] - strides[1]), 0)
        else:
            width_pad = np.max(pool_size[1] - (W % strides[1]), 0)

        pad_top = height_pad // 2
        pad_bottom = height_pad - pad_top
        pad_left = width_pad // 2
        pad_right = width_pad - pad_left
        padding = ((pad_top, pad_bottom), (pad_left, pad_right))
    return np.pad(matrix, padding, mode='constant', constant_values=value)


def max_pooling_2d(image, pool_size=3, strides=1, padding='same'):
    """Returns the maximum pooled value of an image.

    # Arguments
        image: Array.
        poolsize: Int or list of len 2. Window size for each pool
        padding: String. Type of padding
    """
    if not isinstance(strides, int):
        strides = strides[0]
    if not isinstance(pool_size, int):
        pool_size = pool_size[0]

    if padding == 'valid':
        max_image = np.zeros((image.shape[0] - pool_size + 1,
                              image.shape[1] - pool_size + 1))
    if padding == 'same':
        max_image = np.zeros_like(image)

    image = pad_matrix(image, pool_size, strides, padding)
    H, W = image.shape[:2]
    for y in range(0, H - pool_size + 1, strides):
        for x in range(0, W - pool_size + 1, strides):
            max_image[y][x] = np.max(image[y:y + pool_size, x:x + pool_size])
    return max_image


def predict(x, model, preprocess=None, postprocess=None):
    """Preprocess, predict and postprocess input.
    # Arguments
        x: Input to model
        model: Callable i.e. Keras model.
        preprocess: Callable, used for preprocessing input x.
        postprocess: Callable, used for postprocessing output of model.

    # Note
        If model outputs a tf.Tensor is converted automatically to numpy array.
    """
    if preprocess is not None:
        x = preprocess(x)
    y = model(x)
    if isinstance(y, tf.Tensor):
        y = y.numpy()
    if postprocess is not None:
        y = postprocess(y)
    return y



import numpy as np


def to_affine_matrix(rotation_matrix, translation):
    """Builds affine matrix from rotation matrix and translation vector.

    # Arguments
        rotation_matrix: Array (3, 3). Representing a rotation matrix.
        translation: Array (3). Translation vector.

    # Returns
        Array (4, 4) representing an affine matrix.
    """
    if len(translation) != 3:
        raise ValueError('Translation should be of lenght 3')
    if rotation_matrix.shape != (3, 3):
        raise ValueError('Rotation matrix should be of shape (3, 3)')
    translation = translation.reshape(3, 1)
    affine_top = np.concatenate([rotation_matrix, translation], axis=1)
    affine_row = np.array([[0.0, 0.0, 0.0, 1.0]])
    affine_matrix = np.concatenate([affine_top, affine_row], axis=0)
    return affine_matrix


def to_affine_matrices(rotations, translations):
    """Construct affine matrices for rotation matrices vector and
       translation vector.

    # Arguments
        ratations: Rotation matrix vector [N, 3, 3].
        translations: Translation vector [N, 3].

    # Returns
        Transformation matrix [N, 4, 4]
    """
    affine_matrices = []
    for rotation, translation in zip(rotations, translations):
        transformation = to_affine_matrix(rotation, translation)
        affine_matrices.append(transformation)
    return np.array(affine_matrices)



import cv2
import numpy as np

from ..backend.image import resize_image, convert_color_space, show_image
from ..backend.image import BGR2RGB


class Camera(object):
    """Camera abstract class.
    By default this camera uses the openCV functionality.
    It can be inherited to overwrite methods in case another camera API exists.
    """
    def __init__(self, device_id=0, name='Camera', intrinsics=None,
                 distortion=None):
        # TODO load parameters from camera name. Use ``load`` method.
        self.device_id = device_id
        self.name = name
        self.intrinsics = intrinsics
        self.distortion = None
        self._camera = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def intrinsics(self):
        return self._intrinsics

    @intrinsics.setter
    def intrinsics(self, value):
        if value is None:
            value = np.zeros((4))
        self._intrinsics = value

    @property
    def distortion(self):
        return self._distortion

    @distortion.setter
    def distortion(self, distortion):
        self._distortion = distortion

    def start(self):
        """ Starts capturing device

        # Returns
            Camera object.
        """
        self._camera = cv2.VideoCapture(self.device_id)
        if self._camera is None or not self._camera.isOpened():
            raise ValueError('Unable to open device', self.device_id)
        return self._camera

    def stop(self):
        """ Stops capturing device.
        """
        return self._camera.release()

    def read(self):
        """Reads camera input and returns a frame.

        # Returns
            Image array.
        """
        frame = self._camera.read()[1]
        return frame

    def is_open(self):
        """Checks if camera is open.

        # Returns
            Boolean
        """
        return self._camera.isOpened()

    def calibrate(self):
        raise NotImplementedError

    def save(self, filepath):
        raise NotImplementedError

    def load(self, filepath):
        raise NotImplementedError

    def intrinsics_from_HFOV(self, HFOV=70, image_shape=None):
        """Computes camera intrinsics using horizontal field of view (HFOV).

        # Arguments
            HFOV: Angle in degrees of horizontal field of view.
            image_shape: List of two floats [height, width].

        # Returns
            camera intrinsics array (3, 3).

        # Notes:

                       \           /      ^
                        \         /       |
                         \ lens  /        | w/2
        horizontal field  \     / alpha/2 |
        of view (alpha)____\( )/_________ |      image
                           /( )\          |      plane
                          /     <-- f --> |
                         /       \        |
                        /         \       |
                       /           \      v

                    Pinhole camera model

        From the image above we know that: tan(alpha/2) = w/2f
        -> f = w/2 * (1/tan(alpha/2))

        alpha in webcams and phones is often between 50 and 70 degrees.
        -> 0.7 w <= f <= w
        """
        if image_shape is None:
            self.start()
            height, width = self.read().shape[0:2]
            self.stop()
        else:
            height, width = image_shape[:2]

        focal_length = (width / 2) * (1 / np.tan(np.deg2rad(HFOV / 2.0)))
        intrinsics = np.array([[focal_length, 0, width / 2.0],
                               [0, focal_length, height / 2.0],
                               [0, 0, 1.0]])
        self.intrinsics = intrinsics

    def take_photo(self):
        """Starts camera, reads buffer and returns an image.
        """
        self.start()
        image = self.read()
        # all pipelines start with RGB
        image = convert_color_space(image, BGR2RGB)
        self.stop()
        return image


class VideoPlayer(object):
    """Performs visualization inferences in a real-time video.

    # Properties
        image_size: List of two integers. Output size of the displayed image.
        pipeline: Function. Should take RGB image as input and it should
            output a dictionary with key 'image' containing a visualization
            of the inferences. Built-in pipelines can be found in
            ``paz/processing/pipelines``.

    # Methods
        run()
        record()
    """

    def __init__(self, image_size, pipeline, camera, topic='image'):
        self.image_size = image_size
        self.pipeline = pipeline
        self.camera = camera
        self.topic = topic

    def step(self):
        """ Runs the pipeline process once

        # Returns
            Inferences from ``pipeline``.
        """
        if self.camera.is_open() is False:
            raise ValueError('Camera has not started. Call ``start`` method.')

        frame = self.camera.read()
        if frame is None:
            print('Frame: None')
            return None
        # all pipelines start with an RGB image
        frame = convert_color_space(frame, BGR2RGB)
        return self.pipeline(frame)

    def run(self):
        """Opens camera and starts continuous inference using ``pipeline``,
        until the user presses ``q`` inside the opened window.
        """
        self.camera.start()
        while True:
            output = self.step()
            if output is None:
                continue
            image = resize_image(output[self.topic], tuple(self.image_size))
            show_image(image, 'inference', wait=False)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        self.camera.stop()
        cv2.destroyAllWindows()

    def record(self, name='video.avi', fps=20, fourCC='XVID'):
        """Opens camera and records continuous inference using ``pipeline``.

        # Arguments
            name: String. Video name. Must include the postfix .avi.
            fps: Int. Frames per second.
            fourCC: String. Indicates the four character code of the video.
            e.g. XVID, MJPG, X264.
        """
        self.camera.start()
        fourCC = cv2.VideoWriter_fourcc(*fourCC)
        writer = cv2.VideoWriter(name, fourCC, fps, self.image_size)
        while True:
            output = self.step()
            if output is None:
                continue
            image = resize_image(output['image'], tuple(self.image_size))
            show_image(image, 'inference', wait=False)
            writer.write(image)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

        self.camera.stop()
        writer.release()
        cv2.destroyAllWindows()

    def record_from_file(self, video_file_path, name='video.avi',
                         fps=20, fourCC='XVID'):
        """Load video and records continuous inference using ``pipeline``.

        # Arguments
            video_file_path: String. Path to the video file.
            name: String. Output video name. Must include the postfix .avi.
            fps: Int. Frames per second.
            fourCC: String. Indicates the four character code of the video.
            e.g. XVID, MJPG, X264.
        """

        fourCC = cv2.VideoWriter_fourcc(*fourCC)
        writer = cv2.VideoWriter(name, fourCC, fps, self.image_size)

        video = cv2.VideoCapture(video_file_path)
        if (video.isOpened() is False):
            print("Error opening video  file")

        while video.isOpened():
            is_frame_received, frame = video.read()
            if not is_frame_received:
                print("Frame not received. Exiting ...")
                break
            if is_frame_received is True:
                output = self.pipeline(frame)
                if output is None:
                    continue
                image = resize_image(output['image'], tuple(self.image_size))
                show_image(image, 'inference', wait=False)
                writer.write(image)
                if cv2.waitKey(1) & 0xFF == ord('q'):
                    break

        writer.release()
        cv2.destroyAllWindows()



from ..abstract import SequentialProcessor, Processor
from .. import processors as pr

from .image import AugmentImage


class RenderTwoViews(Processor):
    """Renders two views along with their transformations.

    # Arguments
        renderer: A class with a method ``render`` that outputs
            two lists. The first list contains two numpy arrays
            representing the images e.g. ``(image_A, image_B)`` each
            of shape ``[H, W, 3]``.
            The other list contains three numpy arrays representing the
            transformations from the origin to the cameras and the
            two alpha channels of both images e.g.
            ``[matrices, alpha_channel_A, alpha_channel_B]``.
            ``matrices`` is a numpy array of shape ``(4, 4 * 4)``.
            Each row is a matrix of ``4 x 4`` representing the following
            transformations respectively: ``world_to_A``, ``world_to_B``,
            ``A_to_world`` and  ``B_to_world``.
            The shape of each ``alpha_channel`` should be ``[H, W]``.
    """
    def __init__(self, renderer):
        super(RenderTwoViews, self).__init__()
        self.render = pr.Render(renderer)

        self.preprocess_image = SequentialProcessor()
        self.preprocess_image.add(pr.ConvertColorSpace(pr.RGB2BGR))
        self.preprocess_image.add(pr.NormalizeImage())

        self.preprocess_alpha = SequentialProcessor()
        self.preprocess_alpha.add(pr.NormalizeImage())
        self.concatenate = pr.Concatenate(-1)

    def call(self):
        data = self.render()
        image_A = self.preprocess_image(data['image_A'])
        image_B = self.preprocess_image(data['image_B'])
        alpha_A = self.preprocess_alpha(data['alpha_A'])
        alpha_B = self.preprocess_alpha(data['alpha_B'])
        alpha_channels = self.concatenate([alpha_A, alpha_B])
        matrices = data['matrices']
        return image_A, image_B, matrices, alpha_channels


class RandomizeRenderedImage(SequentialProcessor):
    """Performs alpha blending and data-augmentation to an image and
        it's alpha channel.
    image_paths: List of strings indicating the paths to the images used for
        the background.
    num_occlusions: Int. number of occlusions to be added to the image.
    max_radius_scale: Float between [0, 1] indicating the maximum radius in
        scale of the image size.
    """
    def __init__(self, image_paths, num_occlusions=1, max_radius_scale=0.5):
        super(RandomizeRenderedImage, self).__init__()
        self.add(pr.ConcatenateAlphaMask())
        self.add(pr.BlendRandomCroppedBackground(image_paths))
        for arg in range(num_occlusions):
            self.add(pr.AddOcclusion(max_radius_scale))
        self.add(pr.RandomImageBlur())
        self.add(AugmentImage())



class Loader(object):
    """Abstract class for loading a dataset.

    # Arguments
        path: String. Path to data.
        split: String. Dataset split e.g. traing, val, test.
        class_names: List of strings. Label names of the classes.
        name: String. Dataset name.

    # Properties
        name: Str.
        path: Str.
        split: Str or Flag.
        class_names: List of strings.
        num_classes: Int.

    # Methods
        load_data()
    """
    def __init__(self, path, split, class_names, name):
        self.path = path
        self.split = split
        self.class_names = class_names
        self.name = name

    def load_data(self):
        """Abstract method for loading dataset.

        # Returns
            dictionary containing absolute image paths as keys, and
            ground truth vectors as values.
        """
        raise NotImplementedError()

    # Name of the dataset (VOC2007, COCO, OpenImagesV4, etc)
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    #  Path to the dataset, ideally loaded from a configuration file.
    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        self._path = path

    # Kind of split to use, either train, validation, test, or trainval.
    @property
    def split(self):
        return self._split

    @split.setter
    def split(self, split):
        self._split = split

    # List of class names to train/test.
    @property
    def class_names(self):
        return self._class_names

    @class_names.setter
    def class_names(self, class_names):
        # assert type(class_names) == list
        self._class_names = class_names

    @property
    def num_classes(self):
        if isinstance(self.class_names, list):
            return len(self.class_names)
        else:
            raise ValueError('class names are not a list')



import numpy as np
from ..backend.boxes import compute_ious
from ..backend.image import load_image


def compute_matches(dataset, detector, class_to_arg, iou_thresh=0.5):
    """
    Arguments:
        dataset: List of dictionaries containing 'image' as key and a
            numpy array representing an image as value.
        detector : Function for performing inference
        class_to_arg: Dict. of class names and their id
        iou_thresh (float): A prediction is correct if its Intersection over
            Union with the ground truth is above this value..

    Returns:
        num_positives: Dict. containing number of positives for each class
        score: Dict. containing matching scores of boxes for each class
        match: Dict. containing match/non-match info of boxes in each class
    """
    # classes_count = len(np.unique(np.concatenate(ground_truth_class_args)))
    num_classes = len(class_to_arg)
    num_positives = {label_id: 0 for label_id in range(1, num_classes + 1)}
    score = {label_id: [] for label_id in range(1, num_classes + 1)}
    match = {label_id: [] for label_id in range(1, num_classes + 1)}
    for sample in dataset:
        # obtaining ground truths
        ground_truth_boxes = np.array(sample['boxes'][:, :4])
        ground_truth_class_args = np.array(sample['boxes'][:, 4])
        if 'difficulties' in sample.keys():
            difficulties = np.array(sample['difficulties'])
        else:
            difficulties = None
        # obtaining predictions
        image = load_image(sample['image'])
        results = detector(image)
        predicted_boxes, predicted_class_args, predicted_scores = [], [], []
        for box2D in results['boxes2D']:
            predicted_scores.append(box2D.score)
            predicted_class_args.append(class_to_arg[box2D.class_name])
            predicted_boxes.append(list(box2D.coordinates))
        predicted_boxes = np.array(predicted_boxes, dtype=np.float32)
        predicted_class_args = np.array(predicted_class_args)
        predicted_scores = np.array(predicted_scores, dtype=np.float32)
        # setting difficulties to ``Easy`` if they are None
        if difficulties is None:
            difficulties = np.zeros(len(ground_truth_boxes), dtype=bool)
        # iterating over each class present in the image
        class_args = np.concatenate(
            (predicted_class_args, ground_truth_class_args))
        class_args = np.unique(class_args).astype(int)
        for class_arg in class_args:
            # masking predictions by class
            class_mask = class_arg == predicted_class_args
            class_predicted_boxes = predicted_boxes[class_mask]
            class_predicted_scores = predicted_scores[class_mask]
            # sort score from maximum to minimum for masked predictions
            sorted_args = class_predicted_scores.argsort()[::-1]
            class_predicted_boxes = class_predicted_boxes[sorted_args]
            class_predicted_scores = class_predicted_scores[sorted_args]
            # masking ground truths by class
            class_mask = class_arg == ground_truth_class_args
            class_ground_truth_boxes = ground_truth_boxes[class_mask]
            class_difficulties = difficulties[class_mask]
            # the number of positives equals the number of easy boxes
            num_easy = np.logical_not(class_difficulties).sum()
            num_positives[class_arg] = num_positives[class_arg] + num_easy
            # add all predicted scores to scores
            score[class_arg].extend(class_predicted_scores)
            # if not predicted boxes for this class continue
            if len(class_predicted_boxes) == 0:
                continue
            # if not ground truth boxes continue but add zeros as matches
            if len(class_ground_truth_boxes) == 0:
                match[class_arg].extend((0,) * len(class_predicted_boxes))
                continue

            # evaluation on VOC follows integer typed bounding boxes.
            class_predicted_boxes = class_predicted_boxes.copy()
            class_predicted_boxes[:, 2:] = (
                class_predicted_boxes[:, 2:] + 1)
            class_ground_truth_boxes = class_ground_truth_boxes.copy()
            class_ground_truth_boxes[:, 2:] = (
                class_ground_truth_boxes[:, 2:] + 1)

            ious = compute_ious(
                class_predicted_boxes, class_ground_truth_boxes)
            ground_truth_args = ious.argmax(axis=1)
            # set -1 if there is no matching ground truth
            ground_truth_args[ious.max(axis=1) < iou_thresh] = -1
            selected = np.zeros(len(class_ground_truth_boxes), dtype=bool)
            for ground_truth_arg in ground_truth_args:
                if ground_truth_arg >= 0:
                    if class_difficulties[ground_truth_arg]:
                        match[class_arg].append(-1)
                    else:
                        if not selected[ground_truth_arg]:
                            match[class_arg].append(1)
                        else:
                            match[class_arg].append(0)
                    selected[ground_truth_arg] = True
                else:
                    match[class_arg].append(0)
    return num_positives, score, match


def calculate_relevance_metrics(num_positives, scores, matches):
    """Calculates precision and recall.
    Arguments:
        num_positives: Dict. with number of positives for each class
        scores: Dict. with matching scores of boxes for each class
        matches: Dict. wth match/non-match info for boxes for each class
    Returns:
        precision: Dict. with precision values per class
        recall : Dict. with recall values per class
    """
    num_classes = max(num_positives.keys()) + 1
    precision, recall = [None] * num_classes, [None] * num_classes
    for class_arg in num_positives.keys():
        class_positive_matches = np.array(matches[class_arg], dtype=np.int8)
        class_scores = np.array(scores[class_arg])
        order = class_scores.argsort()[::-1]
        class_positive_matches = class_positive_matches[order]
        true_positives = np.cumsum(class_positive_matches == 1)
        false_positives = np.cumsum(class_positive_matches == 0)
        precision[class_arg] = (
            true_positives / (false_positives + true_positives))
        if num_positives[class_arg] > 0:
            recall[class_arg] = true_positives / num_positives[class_arg]
    return precision, recall


def calculate_average_precisions(precision, recall, use_07_metric=False):
    """Calculate average precisions based based on PASCAL VOC evaluation
    Arguments:
        num_positives: Dict. with number of positives for each class
        scores: Dict. with matching scores of boxes for each class
        matches: Dict. wth match/non-match info for boxes for each class
    Returns:
    """

    num_classes = len(precision)
    average_precisions = np.empty(num_classes)
    for class_arg in range(num_classes):
        if precision[class_arg] is None or recall[class_arg] is None:
            average_precisions[class_arg] = np.nan
            continue

        if use_07_metric:
            # 11 point metric
            average_precisions[class_arg] = 0
            for t in np.arange(0., 1.1, 0.1):
                if np.sum(recall[class_arg] >= t) == 0:
                    p_interpolation = 0
                else:
                    p_interpolation = np.max(
                        np.nan_to_num(
                            precision[class_arg]
                        )[recall[class_arg] >= t]
                    )
                average_precision_class = average_precisions[class_arg]
                average_precision_class = (average_precision_class +
                                           (p_interpolation / 11))
                average_precisions[class_arg] = average_precision_class

        else:
            # first append sentinel values at the end
            average_precision = np.concatenate(
                ([0], np.nan_to_num(precision[class_arg]), [0]))
            average_recall = np.concatenate(([0], recall[class_arg], [1]))

            average_precision = np.maximum.accumulate(
                average_precision[::-1])[::-1]

            # to calculate area under PR curve, look for points
            # where X axis (recall) changes value
            recall_change_arg = np.where(
                average_recall[1:] != average_recall[:-1])[0]

            # and sum (\Delta recall) * precision
            average_precisions[class_arg] = np.sum(
                (average_recall[recall_change_arg + 1] -
                 average_recall[recall_change_arg]) *
                average_precision[recall_change_arg + 1])
    return average_precisions


def evaluateMAP(detector, dataset, class_to_arg, iou_thresh=0.5,
                use_07_metric=False):
    """Calculate average precisions based on evaluation code of PASCAL VOC.
    Arguments:
        dataset: List of dictionaries containing 'image' as key and a
            numpy array representing an image as value.
        detector : Function for performing inference
        class_to_arg: Dict. of class names and their id
        iou_thresh: Float indicating intersection over union threshold for
            assigning a prediction as correct.
    # Returns:
    """
    positives, score, match = compute_matches(
        dataset, detector, class_to_arg, iou_thresh)
    precision, recall = calculate_relevance_metrics(positives, score, match)
    average_precisions = calculate_average_precisions(
        precision, recall, use_07_metric)
    return {'ap': average_precisions, 'map': np.nanmean(average_precisions)}



import numpy as np
from paz import processors as pr


class GetHeatmapsAndTags(pr.Processor):
    """Get Heatmaps and Tags from the model output.
    # Arguments
        model: Model weights trained on HigherHRNet model.
        flipped_keypoint_order: List of length 17 (number of keypoints).
            Flipped list of keypoint order.
        data_with_center: Boolean. True is the model is trained using the
            center.
        image: Numpy array. Input image of shape (H, W)

    # Returns
        heatmaps: Numpy array of shape (1, num_keypoints, H, W)
        Tags: Numpy array of shape (1, num_keypoints, H, W)
    """
    def __init__(self, model, flipped_keypoint_order, with_flip,
                 data_with_center, scale_output=True, axes=[0, 3, 1, 2]):
        super(GetHeatmapsAndTags, self).__init__()
        self.with_flip = with_flip
        self.predict = pr.SequentialProcessor(
            [pr.Predict(model), pr.TransposeOutput(axes), pr.ScaleOutput(2)])
        self.get_heatmaps = pr.GetHeatmaps(flipped_keypoint_order)
        self.get_tags = pr.GetTags(flipped_keypoint_order)
        self.postprocess = pr.SequentialProcessor()
        if data_with_center:
            self.postprocess.add(pr.RemoveLastElement())
        if scale_output:
            self.postprocess.add(pr.ScaleOutput(2, full_scaling=True))

    def call(self, image):
        outputs = self.predict(image)
        heatmaps = self.get_heatmaps(outputs, with_flip=False)
        tags = self.get_tags(outputs, with_flip=False)
        if self.with_flip:
            outputs = self.predict(np.flip(image, [2]))
            heatmaps_flip = self.get_heatmaps(outputs, self.with_flip)
            tags_flip = self.get_tags(outputs, self.with_flip)
            heatmaps = [heatmaps, heatmaps_flip]
            tags = [tags, tags_flip]
        heatmaps = self.postprocess(heatmaps)
        tags = self.postprocess(tags)
        return heatmaps, tags



import os
from glob import glob
import json

import numpy as np
from tensorflow.keras.utils import Progbar

from ..abstract import Loader
from .utils import get_class_names


class FAT(Loader):
    """ Dataset loader for the falling things dataset (FAT).

    # Arguments
        path: String indicating full path to dataset
            e.g. /home/user/fat/
        split: String determining the data split to load.
            e.g. `train`, `val` or `test`
        class_names: `all` or list. If list it should contain as elements
            strings indicating each class name.

    # References
        - [Deep Object Pose
            Estimation (DOPE)](https://github.com/NVlabs/Deep_Object_Pose)
    """
    # TODO: Allow selection of class_names.
    def __init__(self, path, split='train', class_names='all'):
        if class_names == 'all':
            class_names = get_class_names('FAT')
        self.class_to_arg = dict(
            zip(class_names, list(range(len(class_names)))))

        super(FAT, self).__init__(path, split, class_names, 'FAT')

    def load_data(self):
        scene_names = glob(self.path + 'mixed/*')
        image_paths, label_paths = [], []
        for scene_name in scene_names:
            scene_image_paths, scene_label_paths = [], []
            for image_side in ['left', 'right']:
                image_names = glob(scene_name + '/*%s.jpg' % image_side)
                side_image_paths = sorted(image_names, key=self._base_number)
                label_names = glob(scene_name + '/0*%s.json' % image_side)
                side_label_paths = sorted(label_names, key=self._base_number)
                scene_image_paths = scene_image_paths + side_image_paths
                scene_label_paths = scene_label_paths + side_label_paths
            image_paths = image_paths + scene_image_paths
            label_paths = label_paths + scene_label_paths

        self.data = []
        progress_bar = Progbar(len(image_paths))
        for sample_arg, sample in enumerate(zip(image_paths, label_paths)):
            image_path, label_path = sample
            if not self._valid_name_match(image_path, label_path):
                raise ValueError('Invalid name match:', image_path, label_path)
            boxes = self._extract_boxes(label_path)
            if boxes is None:
                continue
            self.data.append({'image': image_path, 'boxes': boxes})
            progress_bar.update(sample_arg + 1)
        return self.data

    def _extract_boxes(self, json_filename):
        json_data = json.load(open(json_filename, 'r'))
        num_objects = len(json_data['objects'])
        if num_objects == 0:
            return None
        box_data = np.zeros((num_objects, 5))
        for object_arg, object_data in enumerate(json_data['objects']):
            bounding_box = object_data['bounding_box']
            y_min, x_min = bounding_box['top_left']
            y_max, x_max = bounding_box['bottom_right']
            x_min, y_min = x_min / 960., y_min / 540.
            x_max, y_max = x_max / 960., y_max / 540.
            box_data[object_arg, :4] = x_min, y_min, x_max, y_max
            class_name = object_data['class'][:-4]
            box_data[object_arg, -1] = self.class_to_arg[class_name]
        return box_data

    def _base_number(self, filename):
        order = os.path.basename(filename)
        order = order.split('.')[0]
        order = float(order)
        return order

    def _valid_name_match(self, image_path, label_path):
        image_name = os.path.basename(image_path)
        label_name = os.path.basename(label_path)
        return image_name[:-3] == label_name[:-4]



from ..abstract import Processor
from ..backend.groups import rotation_vector_to_quaternion
from ..backend.groups import rotation_vector_to_rotation_matrix
from ..backend.groups import to_affine_matrix


class RotationVectorToQuaternion(Processor):
    """Transforms rotation vector into quaternion.
    """
    def __init__(self):
        super(RotationVectorToQuaternion, self).__init__()

    def call(self, rotation_vector):
        quaternion = rotation_vector_to_quaternion(rotation_vector)
        return quaternion


class RotationVectorToRotationMatrix(Processor):
    """Transforms rotation vector into a rotation matrix.
    """
    def __init__(self):
        super(RotationVectorToRotationMatrix, self).__init__()

    def call(self, rotation_vector):
        return rotation_vector_to_rotation_matrix(rotation_vector)


class ToAffineMatrix(Processor):
    """Builds affine matrix from a rotation matrix and a translation vector.
    """
    def __init__(self):
        super(ToAffineMatrix, self).__init__()

    def call(self, rotation_matrix, translation):
        affine_matrix = to_affine_matrix(rotation_matrix, translation)
        return affine_matrix



import numpy as np

from ..abstract import Processor

from ..backend.boxes import flip_left_right
from ..backend.boxes import to_image_coordinates
from ..backend.boxes import to_normalized_coordinates
from ..backend.boxes import compute_iou
from ..backend.image import warp_affine
from ..backend.image import translate_image
from ..backend.image import sample_scaled_translation
from ..backend.image import get_rotation_matrix
from ..backend.image import calculate_image_center
from ..backend.keypoints import translate_keypoints
from ..backend.keypoints import rotate_point2D
from ..backend.standard import resize_with_same_aspect_ratio
from ..backend.standard import get_transformation_scale


class RandomFlipBoxesLeftRight(Processor):
    """Flips image and implemented labels horizontally.
    """
    def __init__(self):
        super(RandomFlipBoxesLeftRight, self).__init__()

    def call(self, image, boxes):
        if np.random.randint(0, 2):
            boxes = flip_left_right(boxes, image.shape[1])
            image = image[:, ::-1]
        return image, boxes


class ToImageBoxCoordinates(Processor):
    """Convert normalized box coordinates to image-size box coordinates.
    """
    def __init__(self):
        super(ToImageBoxCoordinates, self).__init__()

    def call(self, image, boxes):
        boxes = to_image_coordinates(boxes, image)
        return image, boxes


class ToNormalizedBoxCoordinates(Processor):
    """Convert image-size box coordinates to normalized box coordinates.
    """
    def __init__(self):
        super(ToNormalizedBoxCoordinates, self).__init__()

    def call(self, image, boxes):
        boxes = to_normalized_coordinates(boxes, image)
        return image, boxes


class RandomSampleCrop(Processor):
    """Crops image while adjusting the normalized corner form
    bounding boxes.

    # Arguments
        probability: Float between ''[0, 1]''.
    """
    def __init__(self, probability=0.50, max_trials=50):
        self.probability = probability
        self.max_trials = max_trials
        self.jaccard_min_max = (
            None,
            (0.1, np.inf),
            (0.3, np.inf),
            (0.7, np.inf),
            (0.9, np.inf),
            (-np.inf, np.inf))

    def call(self, image, boxes):

        if self.probability < np.random.rand():
            return image, boxes

        labels = boxes[:, -1:]
        boxes = boxes[:, :4]
        H_original, W_original = image.shape[:2]

        mode = np.random.randint(0, len(self.jaccard_min_max), 1)[0]
        if self.jaccard_min_max[mode] is not None:
            min_iou, max_iou = self.jaccard_min_max[mode]
            for trial_arg in range(self.max_trials):
                W = np.random.uniform(0.3 * W_original, W_original)
                H = np.random.uniform(0.3 * H_original, H_original)
                aspect_ratio = H / W
                if (aspect_ratio < 0.5) or (aspect_ratio > 2):
                    continue
                x_min = np.random.uniform(W_original - W)
                y_min = np.random.uniform(H_original - H)
                x_max = int(x_min + W)
                y_max = int(y_min + H)
                x_min = int(x_min)
                y_min = int(y_min)

                image_crop_box = np.array([x_min, y_min, x_max, y_max])
                overlap = compute_iou(image_crop_box, boxes)
                if ((overlap.max() < min_iou) or (overlap.min() > max_iou)):
                    continue

                centers = (boxes[:, :2] + boxes[:, 2:]) / 2.0
                centers_above_x_min = x_min < centers[:, 0]
                centers_above_y_min = y_min < centers[:, 1]
                centers_below_x_max = x_max > centers[:, 0]
                centers_below_y_max = y_max > centers[:, 1]
                mask = (centers_above_x_min * centers_above_y_min *
                        centers_below_x_max * centers_below_y_max)
                if not mask.any():
                    continue

                cropped_image = image[y_min:y_max, x_min:x_max, :].copy()
                masked_boxes = boxes[mask, :].copy()
                masked_labels = labels[mask].copy()
                # should we use the box left and top corner or the crop's
                masked_boxes[:, :2] = np.maximum(masked_boxes[:, :2],
                                                 image_crop_box[:2])
                # adjust to crop (by substracting crop's left,top)
                masked_boxes[:, :2] -= image_crop_box[:2]
                masked_boxes[:, 2:] = np.minimum(masked_boxes[:, 2:],
                                                 image_crop_box[2:])
                # adjust to crop (by substracting crop's left,top)
                masked_boxes[:, 2:] -= image_crop_box[:2]
                return cropped_image, np.hstack([masked_boxes, masked_labels])

        boxes = np.hstack([boxes, labels])
        return image, boxes


class Expand(Processor):
    """Expand image size up to 2x, 3x, 4x and fill values with mean color.
    This transformation is applied with a probability of 50%.

    # Arguments
        max_ratio: Float.
        mean: None/List: If `None` expanded image is filled with
            the image mean.
        probability: Float between ''[0, 1]''.
    """
    def __init__(self, max_ratio=2, mean=None, probability=0.5):
        super(Expand, self).__init__()
        self.max_ratio = max_ratio
        self.mean = mean
        self.probability = probability

    def call(self, image, boxes):
        if self.probability < np.random.rand():
            return image, boxes
        height, width, num_channels = image.shape
        ratio = np.random.uniform(1, self.max_ratio)
        left = np.random.uniform(0, width * ratio - width)
        top = np.random.uniform(0, height * ratio - height)
        expanded_image = np.zeros((int(height * ratio),
                                   int(width * ratio), num_channels),
                                  dtype=image.dtype)

        if self.mean is None:
            expanded_image[:, :, :] = np.mean(image, axis=(0, 1))
        else:
            expanded_image[:, :, :] = self.mean

        expanded_image[int(top):int(top + height),
                       int(left):int(left + width)] = image
        expanded_boxes = boxes.copy()
        expanded_boxes[:, 0:2] = boxes[:, 0:2] + (int(left), int(top))
        expanded_boxes[:, 2:4] = boxes[:, 2:4] + (int(left), int(top))
        return expanded_image, expanded_boxes


class ApplyTranslation(Processor):
    """Applies a translation of image and labels.

    # Arguments
        translation: A list of length two indicating the x,y translation values
        fill_color: List of three integers indicating the
            color values e.g. ''[0, 0, 0]''
    """
    def __init__(self, translation, fill_color=None):
        super(ApplyTranslation, self).__init__()
        self._matrix = np.zeros((2, 3), dtype=np.float32)
        self._matrix[0, 0], self._matrix[1, 1] = 1.0, 1.0
        self.fill_color = fill_color
        self.translation = translation

    @property
    def translation(self):
        return self._translation

    @translation.setter
    def translation(self, translation):
        if translation is None:
            self._translation = None
        elif len(translation) == 2:
            self._translation = translation
            self._matrix[0, 2], self._matrix[1, 2] = translation
        else:
            raise ValueError('Translation should be `None` or have length two')

    def call(self, image, keypoints=None):
        height, width = image.shape[:2]
        if self.fill_color is None:
            fill_color = np.mean(image, axis=(0, 1))
        image = warp_affine(image, self._matrix, fill_color)
        if keypoints is not None:
            keypoints[:, 0] = keypoints[:, 0] + self.translation[0]
            keypoints[:, 1] = keypoints[:, 1] + self.translation[1]
            return image, keypoints
        return image


class RandomTranslation(Processor):
    """Applies a random translation to image and labels

    # Arguments
        delta_scale: List with two elements having the normalized deltas.
            e.g. ''[.25, .25]''.

        fill_color: List of three integers indicating the
            color values e.g. ''[0, 0, 0]''.
    """
    def __init__(
            self, delta_scale=[0.25, 0.25], fill_color=None):
        super(RandomTranslation, self).__init__()
        self.delta_scale = delta_scale
        self.apply_translation = ApplyTranslation(None, fill_color)

    @property
    def delta_scale(self):
        return self._delta_scale

    @delta_scale.setter
    def delta_scale(self, delta_scale):
        x_delta_scale, y_delta_scale = delta_scale
        if (x_delta_scale < 0) or (y_delta_scale < 0):
            raise ValueError('Delta scale values should be a positive scalar')
        self._delta_scale = delta_scale

    def call(self, image):
        height, width = image.shape[:2]
        x_delta_scale, y_delta_scale = self.delta_scale
        x = image.shape[1] * np.random.uniform(-x_delta_scale, x_delta_scale)
        y = image.shape[0] * np.random.uniform(-y_delta_scale, y_delta_scale)
        self.apply_translation.translation = [x, y]
        return self.apply_translation(image)


class RandomKeypointTranslation(Processor):
    """Applies a random translation to image and keypoints.

    # Arguments
        delta_scale: List with two elements having the normalized deltas.
            e.g. ''[.25, .25]''.
        fill_color: ''None'' or List of three integers indicating the
            color values e.g. ''[0, 0, 0]''. If ''None'' mean channel values of
            the image will be calculated as fill values.
        probability: Float between ''[0, 1]''.
    """
    def __init__(self, delta_scale=[.2, .2], fill_color=None, probability=0.5):
        super(RandomKeypointTranslation, self).__init__()
        self.delta_scale = delta_scale
        self.fill_color = fill_color
        self.probability = probability

    @property
    def probability(self):
        return self._probability

    @probability.setter
    def probability(self, value):
        if not (0.0 < value <= 1.0):
            raise ValueError('Probability should be between "[0, 1]".')
        self._probability = value

    @property
    def delta_scale(self):
        return self._delta_scale

    @delta_scale.setter
    def delta_scale(self, delta_scale):
        x_delta_scale, y_delta_scale = delta_scale
        if (x_delta_scale < 0) or (y_delta_scale < 0):
            raise ValueError('Delta scale values should be positive')
        if (x_delta_scale > 1) or (y_delta_scale > 1):
            raise ValueError('Delta scale values should be less than one')
        self._delta_scale = delta_scale

    def _sample_random_translation(self, delta_scale, image_shape):
        x_delta_scale, y_delta_scale = delta_scale
        x = image_shape[1] * np.random.uniform(-x_delta_scale, x_delta_scale)
        y = image_shape[0] * np.random.uniform(-y_delta_scale, y_delta_scale)
        return [x, y]

    def call(self, image, keypoints):
        if self.probability >= np.random.rand():
            shape = image.shape[:2]
            translation = sample_scaled_translation(self.delta_scale, shape)
            if self.fill_color is None:
                fill_color = np.mean(image, axis=(0, 1))
            image = translate_image(image, translation, fill_color)
            keypoints = translate_keypoints(keypoints, translation)
        return image, keypoints


class RandomKeypointRotation(Processor):
    """Randomly rotate an images with its corresponding keypoints.

    # Arguments
        rotation_range: Int. indicating the max and min values in degrees
            of the uniform distribution ''[-range, range]'' from which the
            angles are sampled.
        fill_color: ''None'' or List of three integers indicating the
            color values e.g. ''[0, 0, 0]''. If ''None'' mean channel values of
            the image will be calculated as fill values.
    """
    def __init__(self, rotation_range=30, fill_color=None, probability=0.5):
        super(RandomKeypointRotation, self).__init__()
        self.rotation_range = rotation_range
        self.fill_color = fill_color
        self.probability = probability

    @property
    def probability(self):
        return self._probability

    @probability.setter
    def probability(self, value):
        if not (0.0 < value <= 1.0):
            raise ValueError('Probability should be between "[0, 1]".')
        self._probability = value

    def _calculate_image_center(self, image):
        return (int(image.shape[0] / 2), int(image.shape[1] / 2))

    def _rotate_image(self, image, degrees):
        center = self._calculate_image_center(image)
        matrix = get_rotation_matrix(center, degrees)
        if self.fill_color is None:
            fill_color = np.mean(image, axis=(0, 1))
        return warp_affine(image, matrix, fill_color)

    def _degrees_to_radians(self, degrees):
        # negative sign changes rotation direction to follow openCV convention.
        return - (3.14159 / 180) * degrees

    def _build_rotation_matrix(self, radians):
        return np.array([[np.cos(radians), - np.sin(radians)],
                         [np.sin(radians), + np.cos(radians)]])

    def _rotate_keypoints(self, keypoints, radians, image_center):
        keypoints = keypoints - image_center
        matrix = self._build_rotation_matrix(radians)
        keypoints = np.matmul(matrix, keypoints.T).T
        keypoints = keypoints + image_center
        return keypoints

    def _sample_rotation(self, rotation_range):
        return np.random.uniform(-rotation_range, rotation_range)

    def call(self, image, keypoints):
        if self.probability >= np.random.rand():
            degrees = self._sample_rotation(self.rotation_range)
            image = self._rotate_image(image, degrees)
            center = self._calculate_image_center(image)
            radians = self._degrees_to_radians(degrees)
            keypoints = self._rotate_keypoints(keypoints, radians, center)
        return image, keypoints


class RandomRotation(Processor):
    """Randomly rotate an images

    # Arguments
        rotation_range: Int. indicating the max and min values in degrees
            of the uniform distribution ``[-range, range]`` from which the
            angles are sampled.
        fill_color: ''None'' or List of three integers indicating the
            color values e.g. ``[0, 0, 0]``. If ``None`` mean channel values of
            the image will be calculated as fill values.
        probability: Float between 0 and 1.
    """
    def __init__(self, rotation_range=30, fill_color=None, probability=0.5):
        super(RandomRotation, self).__init__()
        self.rotation_range = rotation_range
        self.fill_color = fill_color
        self.probability = probability

    @property
    def probability(self):
        return self._probability

    @probability.setter
    def probability(self, value):
        if not (0.0 < value <= 1.0):
            raise ValueError('Probability should be between "[0, 1]".')
        self._probability = value

    def _calculate_image_center(self, image):
        return (int(image.shape[0] / 2), int(image.shape[1] / 2))

    def _rotate_image(self, image, degrees):
        center = self._calculate_image_center(image)
        matrix = get_rotation_matrix(center, degrees)
        if self.fill_color is None:
            fill_color = np.mean(image, axis=(0, 1))
        return warp_affine(image, matrix, fill_color)

    def _sample_rotation(self, rotation_range):
        return np.random.uniform(-rotation_range, rotation_range)

    def call(self, image):
        if self.probability >= np.random.rand():
            degrees = self._sample_rotation(self.rotation_range)
            image = self._rotate_image(image, degrees)
        return image


class TranslateImage(Processor):
    """Applies a translation of image.
    The translation is a list of length two indicating the x, y values.

    # Arguments
        fill_color: List of three integers indicating the
            color values e.g. ``[0, 0, 0]``
    """
    def __init__(self, fill_color=None):
        super(TranslateImage, self).__init__()
        self.fill_color = fill_color

    def call(self, image, translation):
        return translate_image(image, translation, self.fill_color)


class GetTransformationSize(Processor):
    """Calculate the transformation size for the imgae.
    The size is tuple of length two indicating the x, y values.

    # Arguments
        image: Numpy array
    """
    def __init__(self, input_size, multiple):
        super(GetTransformationSize, self).__init__()
        self.input_size = input_size
        self.multiple = multiple

    def call(self, image):
        size = resize_with_same_aspect_ratio(image, self.input_size,
                                             self.multiple)
        H, W = image.shape[:2]
        if W < H:
            size[0], size[1] = size[1], size[0]
        return size


class GetTransformationScale(Processor):
    """Calculate the transformation scale for the imgae.
    The scale is a numpy array of size two indicating the
    width and height scale.

    # Arguments
        image: Numpy array
        size: Numpy array of length 2
    """
    def __init__(self, scaling_factor):
        super(GetTransformationScale, self).__init__()
        self.scaling_factor = scaling_factor

    def call(self, image, size):
        scale = get_transformation_scale(image, size, self.scaling_factor)
        H, W = image.shape[:2]
        if W < H:
            scale[0], scale[1] = scale[1], scale[0]
        return scale


class GetSourceDestinationPoints(Processor):
    """Returns the source and destination points for affine transformation.

    # Arguments
        center: Numpy array of shape (2,). Center coordinates of image
        scale: Numpy array of shape (2,). Scale of width and height of image
        size: List of length 2. Size of image
    """
    def __init__(self, scaling_factor):
        super(GetSourceDestinationPoints, self).__init__()
        self.scaling_factor = scaling_factor

    def _calculate_third_point(self, point2D_a, point2D_b):
        difference = point2D_a - point2D_b
        return point2D_a + np.array([-difference[1],
                                     difference[0]], dtype=np.float32)

    def _get_transformation_source_point(self, scale, center):
        scale = scale * self.scaling_factor
        center_W = scale[0] / 2
        direction_vector = rotate_point2D([0, -center_W], 0)
        points = np.zeros((3, 2), dtype=np.float32)
        points[0, :] = center
        points[1, :] = center + direction_vector
        points[2:, :] = self._calculate_third_point(points[0, :], points[1, :])
        return points

    def _get_transformation_destination_point(self, output_size):
        center_W, center_H = np.array(output_size[:2]) / 2
        direction_vector = np.array([0, -center_W], np.float32)
        points = np.zeros((3, 2), dtype=np.float32)
        points[0, :] = [center_W, center_H]
        points[1, :] = np.array([center_W, center_H]) + direction_vector
        points[2:, :] = self._calculate_third_point(points[0, :], points[1, :])
        return points

    def call(self, center, scale, size):
        if not isinstance(scale, np.ndarray) and not isinstance(scale, list):
            scale = np.array([scale, scale])
        source_point = self._get_transformation_source_point(scale, center)
        destination_point = self._get_transformation_destination_point(size)
        return source_point, destination_point


class GetImageCenter(Processor):
    """Calculate the center of the image and add an offset to the center.

    # Arguments
        image: Numpy array
        offset: Float
    """
    def __init__(self, offset=0.5):
        super(GetImageCenter, self).__init__()
        self.offset = offset

    def _add_offset(self, x, offset):
        return (x + offset)

    def call(self, image):
        center_W, center_H = calculate_image_center(image)
        center_W = int(self._add_offset(center_W, self.offset))
        center_H = int(self._add_offset(center_H, self.offset))
        return np.array([center_W, center_H])


class WarpAffine(Processor):
    """Applies an affine transformation to an image

    # Arguments
        image: Numpy array
        transform: Numpy array. Transformation matrix
        size: Numpy array. Transformation size
    """
    def __init__(self):
        super(WarpAffine, self).__init__()

    def call(self, image, transform, size):
        image = warp_affine(image, transform, size=size)
        return image



from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import LeakyReLU
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Permute
from tensorflow.keras.layers import Reshape
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Concatenate
from tensorflow.keras.models import Model

from ..layers import ExpectedDepth
from ..layers import ExpectedValue2D
from ..layers import SubtractScalar


def block(x, num_filters, dilation_rate, alpha, name, kernel_size=(3, 3)):
    x = Conv2D(num_filters, kernel_size, dilation_rate=dilation_rate,
               padding='same', name=name)(x)
    x = BatchNormalization()(x)
    x = LeakyReLU(alpha)(x)
    return x


def KeypointNet2D(input_shape, num_keypoints, filters=64, alpha=0.1):
    """Model for discovering keypoint locations in 2D space, modified from

    # Arguments
        input_shape: List of integers indicating ``[H, W, num_channels]``.
        num_keypoints: Int. Number of keypoints to discover.
        filters: Int. Number of filters used in convolutional layers.
        alpha: Float. Alpha parameter of leaky relu.

    # Returns
        Keras/tensorflow model

    # References
        - [Discovery of Latent 3D Keypoints via End-to-end
            Geometric Reasoning](https://arxiv.org/abs/1807.03146)
    """
    width, height = input_shape[:2]
    base = input_tensor = Input(input_shape, name='image')
    for base_arg, rate in enumerate([1, 1, 2, 4, 8, 16, 1, 2, 4, 8, 16, 1]):
        name = 'conv2D_base-%s' % base_arg
        base = block(base, filters, (rate, rate), alpha, name)

    name = 'uv_volume_features-%s'
    uv_volume = Conv2D(num_keypoints, (3, 3),
                       padding='same', name=name % 0)(base)
    uv_volume = Permute([3, 1, 2], name=name % 1)(uv_volume)
    volume_shape = [num_keypoints, width * height]
    uv_volume = Reshape(volume_shape, name=name % 2)(uv_volume)
    uv_volume = Activation('softmax', name=name % 3)(uv_volume)
    volume_shape = [num_keypoints, width, height]
    uv_volume = Reshape(volume_shape, name='uv_volume')(uv_volume)
    uv = ExpectedValue2D(name='keypoints')(uv_volume)
    model = Model(input_tensor, uv, name='keypointnet2D')
    return model


def KeypointNet(input_shape, num_keypoints, depth=.2, filters=64, alpha=0.1):
    """Keypointnet model for discovering keypoint locations in 3D space

    # Arguments
        input_shape: List of integers indicating ``[H, W, num_channels)``.
        num_keypoints: Int. Number of keypoints to discover.
        depth: Float. Prior depth (centimeters) of keypoints.
        filters: Int. Number of filters used in convolutional layers.
        alpha: Float. Alpha parameter of leaky relu.

    # Returns
        Keras/tensorflow model

    # References
        - [Discovery of Latent 3D Keypoints via End-to-end
            Geometric Reasoning](https://arxiv.org/abs/1807.03146)
    """
    width, height = input_shape[:2]
    base = input_tensor = Input(input_shape, name='image')
    for base_arg, rate in enumerate([1, 1, 2, 4, 8, 16, 1, 2, 4, 8, 16, 1]):
        name = 'conv2D_base-%s' % base_arg
        base = block(base, filters, (rate, rate), alpha, name)

    name = 'uv_volume_features-%s'
    uv_volume = Conv2D(num_keypoints, (3, 3),
                       padding='same', name=name % 0)(base)
    uv_volume = Permute([3, 1, 2], name=name % 1)(uv_volume)
    volume_shape = [num_keypoints, width * height]
    uv_volume = Reshape(volume_shape, name=name % 2)(uv_volume)
    uv_volume = Activation('softmax', name=name % 3)(uv_volume)
    volume_shape = [num_keypoints, width, height]
    uv_volume = Reshape(volume_shape, name='uv_volume')(uv_volume)
    uv = ExpectedValue2D(name='expected_uv')(uv_volume)

    name = 'depth_volume_features-%s'
    depth_volume = Conv2D(num_keypoints, (3, 3),
                          padding='same', name=name % 0)(base)
    depth_volume = SubtractScalar(depth, name=name % 1)(depth_volume)
    depth_volume = Permute([3, 1, 2], name='depth_volume')(depth_volume)
    z = ExpectedDepth(name='expected_z')([depth_volume, uv_volume])
    uvz = Concatenate(axis=-1, name='uvz_points')([uv, z])
    model = Model(input_tensor, [uvz, uv_volume], name='keypointnet')
    return model


def KeypointNetShared(input_shape, num_keypoints, depth, filters, alpha):
    """Keypointnet shared model with two views as input.

    # Arguments
        input_shape: List of integers indicating ``[H, W, num_channels]``.
        num_keypoints: Int. Number of keypoints to discover.
        depth: Float. Prior depth (centimeters) of keypoints.
        filters: Int. Number of filters used in convolutional layers.
        alpha: Float. Alpha parameter of leaky relu.

    # Returns
        Keras/tensorflow model

    # References
        - [Discovery of Latent 3D Keypoints via End-to-end
            Geometric Reasoning](https://arxiv.org/abs/1807.03146)
    """

    model_args = (input_shape, num_keypoints, depth, filters, alpha)
    keypointnet = KeypointNet(*model_args)
    image_A = Input(input_shape, name='image_A')
    image_B = Input(input_shape, name='image_B')
    uvz_A, uv_volume_A = keypointnet(image_A)
    uvz_B, uv_volume_B = keypointnet(image_B)
    uvz_points = Concatenate(axis=1, name='uvz_points-shared')([uvz_A, uvz_B])
    uv_volumes = Concatenate(axis=1, name='uv_volumes-shared')(
        [uv_volume_A, uv_volume_B])
    inputs, outputs = [image_A, image_B], [uvz_points, uv_volumes]
    return Model(inputs, outputs, name='keypointnet-shared')



import numpy as np


def sample_point_in_full_sphere(distance=1.0):
    """Get a point of the top of the unit sphere.

    # Arguments
        distance: Float, indicating distance to origin.

    # Returns
        sphere_point: List of spatial coordinates of a sphere.
    """
    if distance <= 0:
        raise ValueError('distance should be bigger than 0')
    sphere_point = np.random.uniform(-1, 1, size=3)
    return (distance * sphere_point) / np.linalg.norm(sphere_point)


def sample_point_in_top_sphere(distance=1.0):
    """Get a point of the top of the unit sphere.

    # Arguments
        distance: Float, indicating distance to origin.

    # Returns
        sphere_point: List of spatial coordinates of a sphere.
    """
    if distance <= 0:
        raise ValueError('distance should be bigger than 0')
    sphere_point = sample_point_in_full_sphere(distance)
    if sphere_point[1] < 0:
        sphere_point[1] = sphere_point[1] * -1
    return sphere_point


def sample_point_in_sphere(distance, top_only=False):
    """ Samples random points from a sphere

    # Arguments
        distance: Float, indicating distance to origin.

    # Returns:
        List of spatial coordinates of a sphere.

    """
    if distance <= 0:
        raise ValueError('distance should be bigger than 0')
    if top_only:
        sphere_point = sample_point_in_top_sphere(distance)
    else:
        sphere_point = sample_point_in_full_sphere(distance)
    return sphere_point


def random_perturbation(localization, shift):
    """Adds noise to 'localization' vector coordinates.

    # Arguments
        localization: List of 3 floats.
        shift: Float indicating a uniform distribution [-shift, shift].

    # Returns
        perturbed localization
    """
    perturbation = np.random.uniform(-shift, shift, size=3)
    return localization + perturbation


def random_translation(localization, shift):
    """Adds noise to 'localization' vector coordinates.

    # Arguments
        localization: List of 3 floats.
        shift: Float indicating a uniform distribution [-shift, shift].
    # Returns
        perturbed localization
    """
    perturbation = np.zeros((3))
    perturbation[:2] = np.random.uniform(-shift, shift, size=2)
    return localization + perturbation


def get_look_at_transform(camera_position, target_position):
    """Make transformation from target position to camera position
    with orientation looking at the target position.

    # Arguments
        camera_position: Numpy-array of length 3. Camera position.
        target_position: Numpy-array of length 3. Target position.
    """
    camera_direction = camera_position - target_position
    camera_direction = camera_direction / np.linalg.norm(camera_direction)
    world_up = np.array([0.0, 1.0, 0.0])
    camera_right = np.cross(world_up, camera_direction)
    camera_right = camera_right / np.linalg.norm(camera_right)
    camera_up = np.cross(camera_direction, camera_right)
    camera_up = camera_up / np.linalg.norm(camera_up)
    rotation_transform = np.zeros((4, 4))
    rotation_transform[0, :3] = camera_right
    rotation_transform[1, :3] = camera_up
    rotation_transform[2, :3] = camera_direction
    rotation_transform[-1, -1] = 1
    translation_transform = np.eye(4)
    translation_transform[:3, -1] = - camera_position
    look_at_transform = np.matmul(rotation_transform, translation_transform)
    return look_at_transform


def compute_modelview_matrices(camera_origin, world_origin,
                               roll=None, translate=None):
    """Compute model-view matrices from camera to origin and origin to camera.

    # Arguments
        camera_origin: Numpy-array of length 3 determining the camera origin
        world_origin: Numpy-array of length 3 determining the world origin
        roll: `None` or float. If `None` no roll is performed. If float
        value should be between [0, 2*pi)

    # Returns
        Transformation from camera to world and world to camera.
    """
    world_to_camera = get_look_at_transform(camera_origin, world_origin)
    if roll is not None:
        world_to_camera = roll_camera(world_to_camera, roll)
    if translate is not None:
        world_to_camera = translate_camera(world_to_camera, translate)
    camera_to_world = np.linalg.inv(world_to_camera)
    return camera_to_world, world_to_camera


def roll_camera(world_to_camera, angle):
    """ Roll camera coordinate system.

    # Arguments:
        world_to_camera: Numpy array containing the affine transformation.
        max_roll: 'None' or float. If None, the camera is not rolled.
            If float it should be a value between [0, 2*pi)
    """
    angle = np.random.uniform(-angle, angle)
    z_rotation = np.array(
        [[np.cos(angle), -np.sin(angle), 0.],
         [np.sin(angle), +np.cos(angle), 0.],
         [0., 0., 1.]])
    world_to_camera[:3, :3] = np.matmul(z_rotation, world_to_camera[:3, :3])
    return world_to_camera


def translate_camera(world_to_camera, translation):
    """ Translate camera coordinate system in its XY plane.

    # Arguments:
        world_to_camera: Numpy array containing the affine transformation.
        translation: List or array with two inputs.
    """
    translation = np.random.uniform(-translation, translation, 2)
    translation_transform = np.array(
        [[1.0, 0.0, 0.0, translation[0]],
         [0.0, 1.0, 0.0, translation[1]],
         [0.0, 0.0, 1.0, 0.0],
         [0.0, 0.0, 0.0, 1.0]])
    world_to_camera = np.matmul(translation_transform, world_to_camera)
    return world_to_camera


def scale_translation(matrix, scale=10.0):
    """ Changes the scale of the translation vector.
    Used for changing the regression problem to a bigger scale.

    # Arguments:
        matrix: Numpy array of shape [4, 4]
        scale: Float used to multiple all the translation component.

    # Returns:
        Numpy array of shape [4, 4]
    """
    matrix[:3, -1] = matrix[:3, -1] * 10.
    return matrix


def sample_uniformly(value):
    """ Samples from a uniform distribution.

    # Arguments
        values: List or float. If list it must have [min_value, max_value].

    # Returns
        Float
    """
    if isinstance(value, list):
        value = np.random.uniform(value[0], value[1])
    return value


def split_alpha_channel(image):
    """ Splits alpha channel from an RGBD image.

    # Arguments
        image: Numpy array of shape [H, W, 4]

    # Returns
        List of two numpy arrays of shape [H, W, 3] and [H, W]
    """
    image_shape = image.shape
    if len(image_shape) != 3:
        raise ValueError('Invalid image shape')
    if image_shape[-1] != 4:
        raise ValueError('Invalid number of channels')
    return image[..., :3], image[..., 3:4]



class Processor(object):
    """Abstract class for creating a processor unit.

    # Arguments
        name: String indicating name of the processing unit.

    # Methods
        call()

    # Example
    ```python
    class NormalizeImage(Processor):
    def __init__(self):
        super(NormalizeImage, self).__init__()

    def call(self, image):
        return image / 255.0
    ```

    # Why this name?
        Originally PAZ was only meant for pre-processing pipelines that
        included data-augmentation, normalization, etc. However, I found
        out that we could use the same API for post-processing; therefore,
        I thought at the time that ``Processor`` would be adequate to describe
        the capacity of both pre-processing and post-processing.
        Names that I also thought could have worked were: ``Function``,
        ``Functor`` but I didn't want to use those since I thought they could
        also cause confusion. Similarly, in Keras this abstraction is
        interpreted as a ``Layer`` but here I don't think that abstraction
        is adequate. A layer of computation maybe? So after having this
        thoughts swirling around I decided to go with ``Processor``
        and try to be explicit about my mental jugglery hoping the name
        doesn't cause much mental overhead.
    """
    def __init__(self, name=None):
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            name = self.__class__.__name__
        self._name = name

    def call(self, X):
        """Custom user's logic should be implemented here.
        """
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)


class SequentialProcessor(object):
    """Abstract class for creating a sequential pipeline of processors.

    # Arguments
        processors: List of instantiated child classes of ``Processor``
            classes.
        name: String indicating name of the processing unit.

    # Methods
        add()
        remove()
        pop()
        insert()
        get_processor()

    # Example
    ```python
    AugmentImage = SequentialProcessor()
    AugmentImage.add(pr.RandomContrast())
    AugmentImage.add(pr.RandomBrightness())
    augment_image = AugmentImage()

    transformed_image = augment_image(image)
    ```
    """
    def __init__(self, processors=None, name=None):
        self.processors = []
        if processors is not None:
            [self.add(processor) for processor in processors]
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            name = self.__class__.__name__
        self._name = name

    def add(self, processor):
        """Adds a process to the sequence of processes to be applied to input.

        # Arguments
            processor: An instantiated child class of of ``Processor``.
        """
        self.processors.append(processor)

    def __call__(self, *args, **kwargs):
        # first call can take list or dictionary values.
        args = self.processors[0](*args, **kwargs)
        # further calls can be a tuple or single values.
        for processor in self.processors[1:]:
            if isinstance(args, tuple):
                args = processor(*args)
            else:
                args = processor(args)
        return args

    def remove(self, name):
        """Removes processor from sequence

        # Arguments
            name: String indicating the process name
        """
        for processor in self.processors:
            if processor.name == name:
                self.processors.remove(processor)

    def pop(self, index=-1):
        """Pops processor in given index from sequence

        # Arguments
            index: Int.
        """
        return self.processors.pop(index)

    def insert(self, index, processor):
        """Inserts ``processor`` to self.processors queue at ``index``

        # Argument
            index: Int.
            processor: An instantiated child class of of ``Processor``.
        """
        return self.processors.insert(index, processor)

    def get_processor(self, name):
        """Gets processor from sequencer

        # Arguments
            name: String indicating the process name
        """
        for processor in self.processors:
            if processor.name == name:
                return processor



import numpy as np


def get_keypoints_heatmap(heatmaps, num_keypoints, indices=None, axis=1):
    """Extract the heatmaps that only contains the keypoints.

    # Arguments
        heatmaps: Numpy array of shape (1, 2*num_keypoints, H, W)
        num_keypoints: Int.
        indices: List. Indices of the heatmaps to extract.
        axis: Int.

    # Returns
        keypoints: Numpy array of shape (1, num_keypoints, H, W)
    """
    keypoints = np.take(heatmaps, np.arange(num_keypoints), axis)
    if indices is not None:
        keypoints = np.take(keypoints, indices, axis)
    return keypoints


def get_tags_heatmap(heatmaps, num_keypoints, indices=None, axis=1):
    """Extract the heatmaps that only contains the tags.

    # Arguments
        heatmaps: Numpy array of shape (1, 2*num_keypoints, H, W)
        num_keypoints: Int.
        indices: List. Indices of the heatmaps to extract.
        axis: Int.

    # Returns
        tags: Numpy array of shape (1, num_keypoints, H, W)
    """
    n = heatmaps.shape[axis]
    tags = np.take(heatmaps, np.arange(num_keypoints, n), axis)
    if indices is not None:
        tags = np.take(tags, indices, axis)
    return tags


def get_keypoints_locations(indices, image_width):
    """Calculate the location of keypoints in an image.

    # Arguments
        indices: Numpy array. Indices of the keypoints in the heatmap.
        Image width: Int.

    # Returns
        coordinate: Numpy array. locations of keypoints
    """
    x = (indices % image_width).astype(np.int64)
    y = (indices / image_width).astype(np.int64)
    coordinates = np.stack((x, y), axis=3)
    return np.squeeze(coordinates)


def get_top_k_keypoints_numpy(heatmaps, k):
    """Numpy implementation of get_top_k_keypoints from heatmaps.

    # Arguments
        heatmaps: Keypoints heatmaps. Numpy array of shape
                  (1, num_keypoints, H, W)
        k: Int. Maximum number of instances to return.

    # Returns
        values: Numpy array. Value of heatmaps at top k keypoints
        indices: Numpy array. Indices of top k keypoints.
    """
    num_of_objects, num_of_keypoints = heatmaps.shape[:2]
    indices = np.zeros((num_of_objects, num_of_keypoints, k), dtype=np.int)
    values = np.zeros((num_of_objects, num_of_keypoints, k))
    for object_arg in range(num_of_objects):
        for keypoint_arg in range(num_of_keypoints):
            top_k_indices = np.argsort(heatmaps[object_arg][keypoint_arg])[-k:]
            top_k_values = heatmaps[object_arg][keypoint_arg][top_k_indices]
            indices[object_arg][keypoint_arg] = top_k_indices
            values[object_arg][keypoint_arg] = top_k_values
    return np.squeeze(values), indices


def get_valid_detections(detection, detection_thresh):
    """Accept the keypoints whose score is greater than the
       detection threshold.

    # Arguments
        detection: Numpy array. Contains the location, value, and
        tags of the keypoints
        detection_thresh: Float. Detection threshold for the keypoint
    """
    mask = detection[:, 2] > detection_thresh
    valid_detection = detection[mask]
    return valid_detection



import numpy as np


def rotation_vector_to_quaternion(rotation_vector):
    """Transforms rotation vector into quaternion.

    # Arguments
        rotation_vector: Numpy array of shape ``[3]``.

    # Returns
        Numpy array representing a quaternion having a shape ``[4]``.
    """
    theta = np.linalg.norm(rotation_vector)
    rotation_axis = rotation_vector / theta
    half_theta = 0.5 * theta
    norm = np.sin(half_theta)
    quaternion = np.array([
        norm * rotation_axis[0],
        norm * rotation_axis[1],
        norm * rotation_axis[2],
        np.cos(half_theta)])
    return quaternion


def homogenous_quaternion_to_rotation_matrix(quaternion):
    """Transforms quaternion to rotation matrix.

    # Arguments
        quaternion: Array containing quaternion value [q1, q2, q3, w0].

    # Returns
        Rotation matrix [3, 3].

    # Note
        If quaternion is not a unit quaternion the rotation matrix is not
        unitary but still orthogonal i.e. the outputted rotation matrix is
        a scalar multiple of a rotation matrix.
    """
    q1, q2, q3, w0 = quaternion

    r11 = w0**2 + q1**2 - q2**2 - q3**2
    r12 = 2 * ((q1 * q2) - (w0 * q3))
    r13 = 2 * ((w0 * q2) + (q1 * q3))

    r21 = 2 * ((w0 * q3) + (q1 * q2))
    r22 = w0**2 - q1**2 + q2**2 - q3**2
    r23 = 2 * ((q2 * q3) - (w0 * q1))

    r31 = 2 * ((q1 * q3) - (w0 * q2))
    r32 = 2 * ((w0 * q1) + (q2 * q3))
    r33 = w0**2 - q1**2 - q2**2 + q3**2

    rotation_matrix = np.array([[r11, r12, r13],
                                [r21, r22, r23],
                                [r31, r32, r33]])
    return rotation_matrix


def quaternion_to_rotation_matrix(quaternion):
    """Transforms quaternion to rotation matrix.

    # Arguments
        quaternion: Array containing quaternion value [q1, q2, q3, w0].

    # Returns
        Rotation matrix [3, 3].

    # Note
        "If the quaternion "is not a unit quaternion then the homogeneous form
        is still a scalar multiple of a rotation matrix, while the
        inhomogeneous form is in general no longer an orthogonal matrix.
        This is why in numerical work the homogeneous form is to be preferred
        if distortion is to be avoided." [wikipedia](https://en.wikipedia.org/
            wiki/Conversion_between_quaternions_and_Euler_angles)
    """
    matrix = homogenous_quaternion_to_rotation_matrix(quaternion)
    return matrix


def rotation_matrix_to_quaternion(rotation_matrix):
    """Transforms rotation matrix to quaternion.

    # Arguments
        Rotation matrix [3, 3].

    # Returns
        quaternion: Array containing quaternion value [q1, q2, q3, w0].
    """
    rotation_matrix = rotation_matrix[:3, :3]
    trace = np.trace(rotation_matrix)
    w0 = np.sqrt(1.0 + trace) / 2
    q1 = 0.25 * (rotation_matrix[2, 1] - rotation_matrix[1, 2]) / w0
    q2 = 0.25 * (rotation_matrix[0, 2] - rotation_matrix[2, 0]) / w0
    q3 = 0.25 * (rotation_matrix[1, 0] - rotation_matrix[0, 1]) / w0
    quaternion = np.array([q1, q2, q3, w0])
    return quaternion


def get_quaternion_conjugate(quaternion):
    """Estimate conjugate of a quaternion.

    # Arguments
        quaternion: Array containing quaternion value [q1, q2, q3, w0].

    # Returns
        quaternion: Array containing quaternion value [-q1, -q2, -q3, w0].
    """
    q1, q2, q3, w0 = quaternion
    return np.array([-q1, -q2, -q3, w0])


def quaternions_to_rotation_matrices(quaternions):
    """Transform quaternion vectors to rotation matrix vector.

    # Arguments
        quaternions [N, 4].

    # Returns
        Rotated matrices [N, 3, 3]
    """
    rotation_matrices = []
    for quaternion in quaternions:
        rotation_matrix = quaternion_to_rotation_matrix(quaternion)
        rotation_matrices.append(rotation_matrix)
    return np.array(rotation_matrices)



import cv2
import numpy as np
from tensorflow.keras.utils import get_file

WEIGHT_PATH = ('https://raw.githubusercontent.com/opencv/opencv/'
               'master/data/haarcascades/')


class HaarCascadeDetector(object):
    """Haar cascade face detector.

    # Arguments
        path: String. Postfix to default openCV haarcascades XML files, see [1]
            e.g. `eye`, `frontalface_alt2`, `fullbody`
        class_arg: Int. Class label argument.
        scale = Float. Scale for image reduction
        neighbors: Int. Minimum neighbors

    # Reference
        - [Haar
            Cascades](https://github.com/opencv/opencv/tree/master/data/haarcascades)
    """

    def __init__(self, weights='frontalface_default', class_arg=None,
                 scale=1.3, neighbors=5):
        self.weights = weights
        self.name = 'haarcascade_' + weights + '.xml'
        self.url = WEIGHT_PATH + self.name
        self.path = get_file(self.name, self.url, cache_subdir='paz/models')
        self.model = cv2.CascadeClassifier(self.path)
        self.class_arg = class_arg
        self.scale = scale
        self.neighbors = neighbors

    def __call__(self, gray_image):
        """ Detects faces from gray images.

        # Arguments
            gray_image: Numpy array of shape ``(H, W, 2)``.

        # Returns
            Numpy array of shape ``(num_boxes, 4)``.
        """
        if len(gray_image.shape) != 2:
            raise ValueError('Invalid gray image shape:', gray_image.shape)
        args = (gray_image, self.scale, self.neighbors)
        boxes = self.model.detectMultiScale(*args)
        boxes_point_form = np.zeros_like(boxes)
        if len(boxes) != 0:
            boxes_point_form[:, 0] = boxes[:, 0]
            boxes_point_form[:, 1] = boxes[:, 1]
            boxes_point_form[:, 2] = boxes[:, 0] + boxes[:, 2]
            boxes_point_form[:, 3] = boxes[:, 1] + boxes[:, 3]
            if self.class_arg is not None:
                class_args = np.ones((len(boxes_point_form), 1))
                class_args = class_args * self.class_arg
                boxes_point_form = np.hstack((boxes_point_form, class_args))
        return boxes_point_form.astype('int')



from paz import processors as pr
from paz.backend.angles import change_link_order
from paz.datasets import MANOHandJoints
from paz.backend.keypoints import flip_along_x_axis
from paz.backend.groups import quaternions_to_rotation_matrices
from paz.backend.groups import to_affine_matrices
from paz.backend.keypoints import rotate_keypoints3D
from paz.backend.keypoints import compute_orientation_vector
from paz.backend.angles import calculate_relative_angle
from paz.backend.angles import reorder_relative_angles
from paz.backend.angles import is_hand_open
from paz.datasets import MPIIHandJoints
from paz.datasets.CMU_poanoptic import hand_part_arg


class ChangeLinkOrder(pr.Processor):
    """Map data from one config to another.

    # Arguments
        joints: Array
        config1_labels: input joint configuration
        config2_labels: output joint configuration

    # Returns
        Array: joints maped to the config2_labels
    """
    def __init__(self, config1_labels, config2_labels):
        super(ChangeLinkOrder, self).__init__()
        self.config1_labels = config1_labels
        self.config2_labels = config2_labels

    def call(self, joints):
        mapped_joints = change_link_order(joints, self.config1_labels,
                                          self.config2_labels)
        return mapped_joints


class CalculateRelativeAngles(pr.Processor):
    """Compute the realtive joint rotation for the minimal hand joints and map
       it to the output_config kinematic chain form.

    # Arguments
        absolute_quaternions : Array [num_joints, 4].
        Absolute joint angle rotation for the minimal hand joints in
        quaternion representation [q1, q2, q3, w0].

    # Returns
        relative_angles: Array [num_joints, 3].
        Relative joint rotation of the minimal hand joints in compact
        axis angle representation.
    """
    def __init__(self, right_hand=False, input_config=MANOHandJoints,
                 output_config=MPIIHandJoints):
        super(CalculateRelativeAngles, self).__init__()
        output_labels = output_config.labels
        input_labels = input_config.labels
        links_origin = input_config.links_origin
        self.parents = input_config.parents
        self.children = output_config.children
        if right_hand:
            links_origin = flip_along_x_axis(links_origin)
        self.links_orientation = compute_orientation_vector(
            links_origin, self.parents)
        self.quaternions_to_rotations = pr.SequentialProcessor([
            pr.ChangeLinkOrder(output_labels, input_labels),
            quaternions_to_rotation_matrices])
        self.calculate_relative_angle = pr.SequentialProcessor([
            calculate_relative_angle,
            pr.ChangeLinkOrder(input_labels, output_labels)])

    def call(self, absolute_quaternions):
        absolute_rotation = self.quaternions_to_rotations(absolute_quaternions)
        rotated_links_origin = rotate_keypoints3D(
            absolute_rotation, self.links_orientation)
        rotated_links_origin_transform = to_affine_matrices(
            absolute_rotation, rotated_links_origin)
        relative_angles = self.calculate_relative_angle(
            absolute_rotation, rotated_links_origin_transform, self.parents)
        relative_angles = reorder_relative_angles(
            relative_angles, absolute_rotation[0], self.children)
        return relative_angles


class IsHandOpen(pr.Processor):
    """Check is the hand is open by by using the relative angles of the joint.
    # Arguments
        joint_name_to_arg: Dictionary for the joints
        thresh: Float. Threshold value for theta
        relative_angle: Array

    # Returns
        String: Hand is open or closed.
    """
    def __init__(self, joint_name_to_arg=hand_part_arg, thresh=0.4):
        super(IsHandOpen, self).__init__()
        self.joint_name_to_arg = joint_name_to_arg
        self.thresh = thresh

    def call(self, relative_angles):
        return is_hand_open(relative_angles, self.joint_name_to_arg,
                            self.thresh)



from ..abstract import SequentialProcessor
from .. import processors as pr
from . import PreprocessImage
from ..models.classification import MiniXception
from ..datasets import get_class_names
from .keypoints import MinimalHandPoseEstimation


# neutral, happiness, surprise, sadness, anger, disgust, fear, contempt
EMOTION_COLORS = [[255, 0, 0], [45, 90, 45], [255, 0, 255], [255, 255, 0],
                  [0, 0, 255], [0, 255, 255], [0, 255, 0]]


class MiniXceptionFER(SequentialProcessor):
    """Mini Xception pipeline for classifying emotions from RGB faces.

    # Example
        ``` python
        from paz.pipelines import MiniXceptionFER

        classify = MiniXceptionFER()

        # apply directly to an image (numpy-array)
        inference = classify(image)
        ```

     # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``class_names`` and ``scores``.

    # References
       - [Real-time Convolutional Neural Networks for Emotion and
            Gender Classification](https://arxiv.org/abs/1710.07557)

    """
    def __init__(self):
        super(MiniXceptionFER, self).__init__()
        self.classifier = MiniXception((48, 48, 1), 7, weights='FER')
        self.class_names = get_class_names('FER')

        preprocess = PreprocessImage(self.classifier.input_shape[1:3], None)
        preprocess.insert(0, pr.ConvertColorSpace(pr.RGB2GRAY))
        preprocess.add(pr.ExpandDims(0))
        preprocess.add(pr.ExpandDims(-1))
        self.add(pr.Predict(self.classifier, preprocess))
        self.add(pr.CopyDomain([0], [1]))
        self.add(pr.ControlMap(pr.ToClassName(self.class_names), [0], [0]))
        self.add(pr.WrapOutput(['class_name', 'scores']))


class ClassifyHandClosure(SequentialProcessor):
    """Pipeline to classify minimal hand closure status.

    # Example
        ``` python
        from paz.pipelines import ClassifyHandClosure

        classify = ClassifyHandClosure()

        # apply directly to an image (numpy-array)
        inference = classify(image)
        ```

     # Returns
        A function that takes an RGB image and outputs an image with class
        status drawn on it.
    """
    def __init__(self, draw=True, right_hand=False):
        super(ClassifyHandClosure, self).__init__()
        self.add(MinimalHandPoseEstimation(draw, right_hand))
        self.add(pr.UnpackDictionary(['image', 'relative_angles']))
        self.add(pr.ControlMap(pr.IsHandOpen(), [1], [1]))
        self.add(pr.ControlMap(pr.BooleanToTextMessage('OPEN', 'CLOSE'),
                               [1], [1]))
        if draw:
            self.add(pr.ControlMap(pr.DrawText(), [0, 1], [0], {1: 1}))
        self.add(pr.WrapOutput(['image', 'status']))



from ..abstract import Processor


class Render(Processor):
    """Render images and labels.

    # Arguments
        renderer: Object that renders images and labels using a method
            ''render_sample()''.
    """
    def __init__(self, renderer):
        super(Render, self).__init__()
        self.renderer = renderer

    def call(self):
        return self.renderer.render()



from .multi_box_loss import MultiBoxLoss
from .keypointnet_loss import KeypointNetLoss
from .segmentation import DiceLoss
from .segmentation import FocalLoss
from .segmentation import JaccardLoss
from .segmentation import WeightedReconstruction
from .segmentation import WeightedReconstructionWithError



def get_class_names(dataset_name='VOC2007'):
    """Gets label names for the classes of the supported datasets.

    # Arguments
        dataset_name: String. Dataset name. Valid dataset names are:
            VOC2007, VOC2012, COCO and YCBVideo.

    # Returns
       List of strings containing the class names for the dataset given.

    # Raises
        ValueError: in case of invalid dataset name
    """

    if dataset_name in ['VOC2007', 'VOC2012', 'VOC']:

        class_names = ['background', 'aeroplane', 'bicycle', 'bird', 'boat',
                       'bottle', 'bus', 'car', 'cat', 'chair', 'cow',
                       'diningtable', 'dog', 'horse', 'motorbike', 'person',
                       'pottedplant', 'sheep', 'sofa', 'train', 'tvmonitor']

    elif dataset_name == 'COCO':
        class_names = ['background', 'person', 'bicycle', 'car', 'motorcycle',
                       'airplane', 'bus', 'train', 'truck', 'boat',
                       'traffic light', 'fire hydrant', 'stop sign',
                       'parking meter', 'bench', 'bird', 'cat', 'dog',
                       'horse', 'sheep', 'cow', 'elephant', 'bear',
                       'zebra', 'giraffe', 'backpack', 'umbrella',
                       'handbag', 'tie', 'suitcase', 'frisbee', 'skis',
                       'snowboard', 'sports ball', 'kite', 'baseball bat',
                       'baseball glove', 'skateboard', 'surfboard',
                       'tennis racket', 'bottle', 'wine glass',
                       'cup', 'fork', 'knife', 'spoon', 'bowl', 'banana',
                       'apple', 'sandwich', 'orange', 'broccoli', 'carrot',
                       'hot dog', 'pizza', 'donut', 'cake', 'chair', 'couch',
                       'potted plant', 'bed', 'dining table', 'toilet',
                       'tv', 'laptop', 'mouse', 'remote', 'keyboard',
                       'cell phone', 'microwave', 'oven', 'toaster',
                       'sink', 'refrigerator', 'book', 'clock', 'vase',
                       'scissors', 'teddy bear', 'hair drier', 'toothbrush']

    elif dataset_name == 'COCO_EFFICIENTDET':
        class_names = ['person', 'bicycle', 'car', 'motorcycle',
                       'airplane', 'bus', 'train', 'truck', 'boat',
                       'traffic light', 'fire hydrant', '0', 'stop sign',
                       'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse',
                       'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe',
                       '0', 'backpack', 'umbrella', '0', '0', 'handbag', 'tie',
                       'suitcase', 'frisbee', 'skis', 'snowboard',
                       'sports ball', 'kite', 'baseball bat', 'baseball glove',
                       'skateboard', 'surfboard', 'tennis racket', 'bottle',
                       '0', 'wine glass', 'cup', 'fork', 'knife', 'spoon',
                       'bowl', 'banana', 'apple', 'sandwich', 'orange',
                       'broccoli', 'carrot', 'hot dog', 'pizza', 'donut',
                       'cake', 'chair', 'couch', 'potted plant', 'bed', '0',
                       'dining table', '0', '0', 'toilet', '0', 'tv', 'laptop',
                       'mouse', 'remote', 'keyboard', 'cell phone',
                       'microwave', 'oven', 'toaster', 'sink', 'refrigerator',
                       '0', 'book', 'clock', 'vase', 'scissors', 'teddy bear',
                       'hair drier', 'toothbrush']

    elif dataset_name == 'YCBVideo':
        class_names = ['background', '037_scissors', '008_pudding_box',
                       '024_bowl', '005_tomato_soup_can', '007_tuna_fish_can',
                       '010_potted_meat_can', '061_foam_brick', '011_banana',
                       '035_power_drill', '004_sugar_box', '019_pitcher_base',
                       '006_mustard_bottle', '036_wood_block',
                       '009_gelatin_box', '051_large_clamp',
                       '040_large_marker', '003_cracker_box',
                       '025_mug', '052_extra_large_clamp',
                       '021_bleach_cleanser', '002_master_chef_can']

    elif dataset_name == 'FAT':
        class_names = ['background', '037_scissors', '008_pudding_box',
                       '024_bowl', '005_tomato_soup_can', '007_tuna_fish_can',
                       '010_potted_meat_can', '061_foam_brick', '011_banana',
                       '035_power_drill', '004_sugar_box', '019_pitcher_base',
                       '006_mustard_bottle', '036_wood_block',
                       '009_gelatin_box', '051_large_clamp',
                       '040_large_marker', '003_cracker_box',
                       '025_mug', '052_extra_large_clamp',
                       '021_bleach_cleanser', '002_master_chef_can']

    elif dataset_name == 'FERPlus':
        return ['neutral', 'happiness', 'surprise', 'sadness',
                'anger', 'disgust', 'fear', 'contempt']

    elif dataset_name == 'FER':
        return ['angry', 'disgust', 'fear', 'happy',
                'sad', 'surprise', 'neutral']

    elif dataset_name == 'IMDB':
        return ['man', 'woman']

    elif dataset_name == 'CityScapes':
        return ['void', 'flat', 'construction',
                'object', 'nature', 'sky', 'human', 'vehicle']

    else:
        raise ValueError('Invalid dataset', dataset_name)

    return class_names


def get_arg_to_class(class_names):
    """Constructs dictionary from argument to class names.

    # Arguments
        class_names: List of strings containing the class names.

    # Returns
        Dictionary mapping integer to class name.
    """

    return dict(zip(list(range(len(class_names))), class_names))
